import os
import time
import cv2
import gymnasium as gym
import numpy as np
from tqdm import tqdm
from itertools import cycle, islice
import multiprocessing
import json

import egogym
from egogym.wrappers import EpisodeMonitor, get_vlm_wrapper
import egogym.assets.constants as constants

def evaluate_policy(
        task_name="pick",
        policy=None,
        robot="rum",
        action_space="delta",
        num_objs=1,
        num_episodes=100,
        max_steps=50,
        render=False,
        record=False,
        render_freq=0,
        render_size=(299, 224),
        num_envs=1,
        seed=42,
        log_file=None,
        objects_set=None,
        reward_threshold=None,
        use_unprivileged_vlm=None,
        logs_dir="logs"
    ):

    num_envs = max(1, min(num_envs, num_episodes))
    if num_envs > 1 and render_freq > 0:
        print(f"Warning: render_freq={render_freq} is not supported with multiple environments. Setting render_freq=0.")
        render_freq = 0
    task_name = task_name.lower()
    if objects_set is None:
        if task_name == "pick":
            objects_set = constants.lite_pick_objects_set
        elif task_name == "open":
            objects_set = constants.all_open_objects_set
    else:
        objects_set_options = {
            'all_pick': constants.all_pick_objects_set,
            'lite_pick': constants.lite_pick_objects_set,
            'diverse_pick': constants.diverse_pick_objects_set,
            'full_eval': constants.full_eval_objects_set,
            'all_open': constants.all_open_objects_set,
            "all_close": constants.all_close_objects_set,
            'cabinet': constants.cabinet_objects_set,
            'drawer': constants.drawer_objects_set,
        }
        if objects_set in objects_set_options:
            objects_set = objects_set_options[objects_set]
        else:
            raise ValueError(f"Unknown objects_set: {objects_set}")
        
    if reward_threshold is None:
        if task_name == "pick":
            reward_threshold = 0.05
        elif task_name == "open":
            reward_threshold = 0.5

    if not log_file:
        log_file = f"evaluation_{int(time.time())}"

    logs_dir = f"{logs_dir}/{log_file}"
    os.makedirs(logs_dir, exist_ok=True)
    
    eval_args = {
        "task_name": task_name,
        "robot": robot,
        "action_space": action_space,
        "num_objs": num_objs,
        "num_episodes": num_episodes,
        "max_steps": max_steps,
        "render": render,
        "record": record,
        "render_freq": render_freq,
        "render_size": render_size,
        "num_envs": num_envs,
        "seed": seed,
        "log_file": log_file,
        "objects_set": objects_set,
        "reward_threshold": reward_threshold,
    }
    
    with open(os.path.join(logs_dir, "eval_args.json"), "w") as f:
        json.dump(eval_args, f, indent=4)

    if num_envs > 1:
        multiprocessing.set_start_method("spawn", force=True)
        
        expanded_objects = list(islice(cycle(objects_set), num_episodes))
        env_objects = [expanded_objects[i::num_envs] for i in range(num_envs)]
        
        def make_env(env_id):
            def _init():
                env = gym.make(
                    f"Egogym-{task_name.capitalize()}-v0",
                    robot=robot,
                    action_space=action_space,
                    render_mode='human' if render else None,
                    render_size=render_size,
                    num_objs=num_objs,
                    seed=seed + env_id,
                    objects_set=env_objects[env_id]
                )
                if use_unprivileged_vlm is not None:
                    wrapper_cls = get_vlm_wrapper(use_unprivileged_vlm)
                    env = wrapper_cls(env)
                return env
            return _init
        
        env = gym.vector.AsyncVectorEnv(
            [make_env(i) for i in range(num_envs)],
            autoreset_mode=gym.vector.AutoresetMode.DISABLED
        )
        env = EpisodeMonitor(env, logs_dir=logs_dir, record=record, render_freq=render_freq, num_envs=num_envs)
        obs, _ = env.reset()
        env_episode_count = np.zeros(num_envs, dtype=int)
        env_step_count = np.zeros(num_envs, dtype=int)
        env_success_count = 0
        episodes_per_env = np.array([len(env_objects[i]) for i in range(num_envs)])
        
        pbar = tqdm(total=num_episodes, desc="Evaluating | Success Rate: 0.0%")
        
        while np.any(env_episode_count < episodes_per_env):
            actions = policy.get_action(obs)
            inactive = env_episode_count >= episodes_per_env
            actions[inactive] = None

            obs, rewards, terminated, truncated, info = env.step(actions)
            env_step_count += 1
            
            if render:
                ego_exo = np.concatenate([obs["rgb_ego"], obs["rgb_exo"]], axis=2)
                combined_view = np.concatenate(ego_exo, axis=0)
                cv2.imshow("Ego View", cv2.cvtColor(combined_view, cv2.COLOR_RGB2BGR))
                cv2.waitKey(1)
            
            dones = (rewards > reward_threshold) | terminated | (env_step_count >= max_steps)
            done_indices = np.where(dones)[0]

            active_done = done_indices[env_episode_count[done_indices] < episodes_per_env[done_indices]]
            
            if len(active_done) > 0:
                env.log_episodes(active_done)
                
                successes = rewards[active_done] > reward_threshold
                env_success_count += np.sum(successes)
                
                env_episode_count[active_done] += 1
                pbar.update(len(active_done))
                
                total_completed = np.sum(env_episode_count)
                success_rate = (env_success_count / total_completed) * 100 if total_completed > 0 else 0
                pbar.set_description(f"Evaluating | Success Rate: {success_rate:.1f}%")
                
                policy.reset(active_done.tolist())
                
                reset_mask = np.zeros(num_envs, dtype=bool)
                reset_mask[active_done] = True
                obs, _ = env.reset(options={"reset_mask": reset_mask})
                env_step_count[active_done] = 0
            
        pbar.close()
        success_rate = (env_success_count / num_episodes) * 100
        return success_rate
        
    else:

        env = gym.make(
            f"Egogym-{task_name.capitalize()}-v0",
            robot=robot,
            action_space=action_space,
            render_mode='human' if render else None,
            render_size=render_size,
            num_objs=num_objs,
            seed=seed,
            objects_set=objects_set
        )
        if use_unprivileged_vlm is not None:
            wrapper_cls = get_vlm_wrapper(use_unprivileged_vlm)
            env = wrapper_cls(env)
        env = EpisodeMonitor(env, logs_dir=logs_dir, record=record, render_freq=render_freq, num_envs=num_envs)
        
        success_count = 0
        pbar = tqdm(range(num_episodes), desc="Evaluating | Success Rate: 0.0%")
        for episode_idx in pbar:
            obs, _ = env.reset()
            policy.reset()
            
            success = False
            for _ in range(max_steps):
                action = policy.get_action(obs)
                obs, reward, terminated, truncated, _ = env.step(action[0])
                    
                if reward > reward_threshold:
                    success = True
                    break
                if terminated or truncated:
                    break
            
            if success:
                success_count += 1
            
            success_rate = (success_count / (episode_idx + 1)) * 100
            pbar.set_description(f"Evaluating | Success Rate: {success_rate:.1f}%")

    env.close()
    return (success_count / num_episodes) * 100 if num_envs == 1 else success_rate


import logging
from typing import Iterable, Optional, Tuple, Union

import torch
from timm.data.random_erasing import RandomErasing
from torchvision import transforms
from torchvision.transforms import InterpolationMode

DEFAULT_CROP_PCT = 0.875
DEFAULT_CROP_MODE = "center"
IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)
IMAGENET_INCEPTION_MEAN = (0.5, 0.5, 0.5)
IMAGENET_INCEPTION_STD = (0.5, 0.5, 0.5)
IMAGENET_DPN_MEAN = (124 / 255, 117 / 255, 104 / 255)
IMAGENET_DPN_STD = tuple([1 / (0.0167 * 255)] * 3)
OPENAI_CLIP_MEAN = (0.48145466, 0.4578275, 0.40821073)
OPENAI_CLIP_STD = (0.26862954, 0.26130258, 0.27577711)


def create_transform(
    input_size,
    is_training: bool = False,
    scale: Optional[Tuple[float, float]] = None,
    ratio: Optional[Tuple[float, float]] = None,
    hflip: float = 0.5,
    vflip: float = 0.0,
    grayscale: float = 0.0,
    gaussblr: float = 0.0,
    gaussblr_kernel: int = 3,
    gaussblr_sigma: tuple = (1.0, 2.0),
    color_jitter: Optional[Union[Iterable[float], float]] = 0.4,
    interpolation: Union[str, InterpolationMode] = "bilinear",
    mean: Union[Iterable, torch.Tensor] = IMAGENET_DEFAULT_MEAN,
    std: Union[Iterable, torch.Tensor] = IMAGENET_DEFAULT_STD,
    re_prob=0.0,
    re_mode="const",
    re_count=1,
    re_num_splits=0,
    crop_pct=None,
    crop_mode=None,
    *args,
    **kwargs,
):
    if isinstance(input_size, (tuple, list)):
        img_size = input_size[-2:]
    else:
        img_size = input_size
    transform_list = []

    interpolation_mode = InterpolationMode.BILINEAR
    try:
        interpolation_mode = InterpolationMode[interpolation.upper()]
    except AttributeError:
        logging.warning(
            f"Interpolation mode {interpolation} is not recognized, "
            f"using bilinear instead."
        )
    if is_training:
        scale = tuple(scale or (0.08, 1.0))  # default imagenet scale range
        ratio = tuple(ratio or (3.0 / 4.0, 4.0 / 3.0))  # default imagenet ratio range

        transform_list.append(
            transforms.RandomResizedCrop(
                img_size,
                scale,
                ratio,
                antialias=False,
                interpolation=interpolation_mode,
            )
        )
        if hflip > 0.0:
            transform_list.append(transforms.RandomHorizontalFlip(hflip))
        if vflip > 0.0:
            transform_list.append(transforms.RandomVerticalFlip(vflip))

        if color_jitter is not None:
            if isinstance(color_jitter, (list, tuple)):
                assert len(color_jitter) in (
                    3,
                    4,
                ), "expected either 3 or 4 values for color jitter"
            else:
                color_jitter = (float(color_jitter),) * 3
            transform_list.append(transforms.ColorJitter(*color_jitter))

        if grayscale > 0.0:
            transform_list.append(transforms.RandomGrayscale(p=grayscale))
        if gaussblr > 0.0:
            transform_list.append(
                transforms.RandomApply(
                    [
                        transforms.GaussianBlur(
                            (gaussblr_kernel, gaussblr_kernel), gaussblr_sigma
                        )
                    ],
                    p=gaussblr,
                )
            )
        transform_list.append(transforms.Normalize(mean, std))

        if re_prob > 0.0:
            transform_list.append(
                RandomErasing(
                    re_prob,
                    mode=re_mode,
                    max_count=re_count,
                    num_splits=re_num_splits,
                    device="cuda",
                )
            )

    else:
        if crop_pct is None:
            crop_pct = DEFAULT_CROP_PCT
        if crop_mode is None:
            crop_mode = DEFAULT_CROP_MODE
        rescaled_size = (
            int(img_size / crop_pct)
            if isinstance(img_size, (int, float))
            else ([int(size / crop_pct) for size in img_size])
        )
        # TODO (haritheja): we are no longer cropping, but figure this out
        # transform_list.append(
        #     transforms.Resize(rescaled_size, interpolation_mode, antialias=False)
        # )
        # if crop_mode == "center":
        #     transform_list.append(transforms.CenterCrop(img_size))
        # elif crop_mode == "random":
        #     transform_list.append(transforms.RandomCrop(img_size))
        # else:
        #     raise ValueError(f"crop_mode '{crop_mode}' not recognized")
        transform_list.append(transforms.Normalize(mean, std))

    return transforms.Compose(transform_list)



from gymnasium.utils import seeding
import gymnasium as gym
from gymnasium.spaces import Box, Dict
import mujoco
import numpy as np
import cv2

from egogym.embodiments import ROBOT_REGISTRY

class Egogym(gym.Env):
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 30}

    def __init__(self, robot="rum", action_space="delta", render_mode=None, render_size=(960,720), seed=None):
        super().__init__()
        self.model = None
        self.data = None
        self.step_idx = 0
        self.render_mode = render_mode
        self.np_random, _ = seeding.np_random(seed)
    
        self.render_width = render_size[0]
        self.render_height = render_size[1]
        
        self.robot_type = robot
        self.action_space_type = action_space
        
        if self.robot_type not in ROBOT_REGISTRY:
            raise ValueError(f"Unknown robot type '{self.robot_type}'. Available robots: {list(ROBOT_REGISTRY.keys())}")
        
        robot_class = ROBOT_REGISTRY[self.robot_type]
        self.robot = robot_class(action_space=self.action_space_type, np_random=self.np_random)
        
        obs_spaces = {
            "rgb_ego": Box(low=0, high=255, shape=(self.render_height, self.render_width, 3), dtype=np.uint8),
            "rgb_exo": Box(low=0, high=255, shape=(self.render_height, self.render_width, 3), dtype=np.uint8),
            "tcp_pose": Box(low=-np.inf, high=np.inf, shape=(16,), dtype=np.float32),
            "camera_pose": Box(low=-np.inf, high=np.inf, shape=(16,), dtype=np.float32),
            "grasp": Box(low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32),
        }
        obs_spaces.update(self.robot.additional_obs_spaces())
        self.observation_space = Dict(obs_spaces)
        self.action_space = Box(low=-np.inf, high=np.inf, shape=self.robot.action_space_shape(), dtype=np.float32)
        self._mujoco_step_counter = 0
        self._render_freq = 0
        self._render_callback = None
        self._should_render = self.render_mode == 'human'

    def setup_mujoco(self, scene_xml: str):
        if hasattr(self, "model") and self.model is not None:
            del self.model
            del self.data
        self.model = mujoco.MjModel.from_xml_string(scene_xml)
        self.data = mujoco.MjData(self.model)
        self.robot.setup_mj(self.model, self.data)
        self.robot.setup_renderers(self.render_width, self.render_height)

    def env_step(self, nsteps=1):
        for _ in range(nsteps):
            mujoco.mj_step(self.model, self.data)
            self._mujoco_step_counter += 1
            if self._render_freq > 0 and self._render_callback is not None and self._mujoco_step_counter >= self._render_freq:
                self._render_callback()
                self._mujoco_step_counter = 0

    def get_base_obs(self):
        obs = {
            "rgb_ego": self.robot.get_camera_view(self.robot.camera_names[0]),
            "rgb_exo": self.robot.get_camera_view(self.robot.camera_names[-1]),
            "tcp_pose": self.robot.get_tcp_pose().reshape(16).astype(np.float32),
            "camera_pose": self.robot.get_camera_pose().reshape(16).astype(np.float32),
            "grasp": np.array([self.robot.get_grasp()], dtype=np.float32),
        }
        obs.update(self.robot.get_additional_obs())
        return obs

    def render(self):
        rgb_ego = self.robot.get_camera_view(self.robot.camera_names[0])
        rgb_exo = self.robot.get_camera_view(self.robot.camera_names[1])
        rgb_stack = np.concatenate((rgb_ego, rgb_exo), axis=1)
        rgb_stack = cv2.cvtColor(rgb_stack, cv2.COLOR_RGB2BGR)
        cv2.imshow("Egogym", rgb_stack)
        cv2.waitKey(1)

    def close(self):
        if self.render_mode == "human":
            cv2.destroyAllWindows()

    def step(self, action):
        if action is None or np.isnan(action).any():
            return self.get_obs(), 0.0, False, False, {}
        self.robot.step(action, step_fn=self.env_step)
        obs = self.get_obs()
        reward = self.compute_reward()
        terminated = False
        truncated = False
        return obs, reward, terminated, truncated, {}
    
    def enable_sleeping_islands(self):
        try:
            self.model.opt.enableflags |= mujoco.mjtEnableBit.mjENBL_SLEEP
        except AttributeError:
            pass
    
    def settle_env(self, nsteps=200):
        if hasattr(self.robot, 'free_joint_id') and self.robot.free_joint_id is not None: #TODO: check if its floating_gripper instead
            self.data.qpos[:] = np.where((np.arange(len(self.data.qpos)) >= self.robot.free_joint_id) & (np.arange(len(self.data.qpos)) < self.robot.free_joint_id + 6), self.data.qpos, 0.0)


import os
import pandas as pd
import matplotlib.pyplot as plt


def plot_success_rates(logs_dir):
    """
    Plot success rates from all CSV files found recursively in the logs directory.
    
    Args:
        logs_dir: Directory containing subdirectories with CSV files
    """
    success_rates = {}
    
    # Recursively find all CSV files
    for root, dirs, files in os.walk(logs_dir):
        for file in files:
            if file.endswith('.csv'):
                csv_path = os.path.join(root, file)
                
                # Get relative path from logs_dir
                rel_path = os.path.relpath(csv_path, logs_dir)
                
                try:
                    # Read the CSV file
                    df = pd.read_csv(csv_path, sep="\t")
                    
                    # Calculate success rate (assuming max_reward > 0 means success)
                    success_rate = (df["max_reward"] > 0.03).mean() * 100
                    
                    success_rates[rel_path] = success_rate
                    print(f"{rel_path}: {success_rate:.2f}%")
                    
                except Exception as e:
                    print(f"Error processing {rel_path}: {e}")
                    continue
    
    if not success_rates:
        print("No valid CSV files found!")
        return
    
    # Sort by relative path
    sorted_items = sorted(success_rates.items())
    
    folders, rates = zip(*sorted_items)
    
    # Create bar plot
    plt.figure(figsize=(12, 6))
    bars = plt.bar(range(len(folders)), rates, color='steelblue', edgecolor='black')
    
    # Add value labels on top of bars
    for i, bar in enumerate(bars):
        height = bar.get_height()
        plt.text(bar.get_x() + bar.get_width()/2., height,
                f'{height:.1f}%',
                ha='center', va='bottom', fontsize=8)
    
    plt.xticks(range(len(folders)), folders, rotation=45, ha='right', fontsize=8)
    plt.xlabel('Log File', fontsize=12)
    plt.ylabel('Success Rate (%)', fontsize=12)
    plt.title('Success Rate by Log File', fontsize=14, fontweight='bold')
    plt.ylim(0, 100)
    plt.grid(axis='y', alpha=0.3)
    plt.tight_layout()
    
    # Save the plot
    output_path = os.path.join(logs_dir, 'success_rates.png')
    plt.savefig(output_path, dpi=300, bbox_inches='tight')
    print(f"\nPlot saved to: {output_path}")
    
    plt.show()


if __name__ == "__main__":
    import sys
    
    if len(sys.argv) > 1:
        logs_directory = sys.argv[1]
    else:
        # Default to logs folder in the project
        logs_directory = "logs"
    
    if not os.path.exists(logs_directory):
        print(f"Error: Directory '{logs_directory}' not found!")
        sys.exit(1)
    
    plot_success_rates(logs_directory)



import mujoco
from scipy.spatial.transform import Rotation as R
import numpy as np

from egogym.utils import get_pose, include_in_scene, add_xml_to_scene


class Object:
    def __init__(self, name, path=None, func=None, np_random=None):
        self.name = name
        self.english_name = self.name.rsplit("_", 1)[0].replace("_", " ")
        self.path = path
        self.last_set_pose = None
        self.object_xml = None
        self.handle_type = None
        self.object_geoms_ids = None
        self.np_random = np_random
        if func is not None:
            self.object_xml = func(
                np_random=np_random
            )
        self.body_name = f"{self.name}_object"

    def get_pose(self, data):
        return get_pose(data, self.body_name, "body")

    def get_geom_ids(self, model):
        if self.object_geoms_ids is not None:
            return self.object_geoms_ids
        object_geoms_ids = []
        for i in range(model.ngeom):
            geom_bodyid = model.geom_bodyid[i]
            body_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_BODY, geom_bodyid)
            if body_name is not None and f"{self.name}_object" in body_name:
                object_geoms_ids.append(i)
        self.object_geoms_ids = object_geoms_ids
        return object_geoms_ids

    def get_bottom_pose(self, data):
        return get_pose(data, f"{self.name}_bottom_site", "site")

    def get_center_to_bottom_z_distance(self, data):
        bottom_z = self.get_bottom_pose(data)[2, 3]
        center_z = self.get_pose(data)[2, 3]
        return center_z - bottom_z

    def set_pose(self, model, data, pose):
        body_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_BODY, self.body_name)
        model.body_pos[body_id] = pose[0:3, 3]
        model.body_quat[body_id] = R.from_matrix(pose[0:3, 0:3]).as_quat()
        self.last_set_pose = pose.copy()

    def add_to_scene_xml(self, scene_xml):
        if self.object_xml is not None:
            return add_xml_to_scene(scene_xml, self.object_xml)
        elif self.path is not None:
            return include_in_scene(scene_xml, self.path)
        else:
            raise ValueError("Either func or path must be provided")


class ArticulableObject(Object):
    def __init__(self, name, path=None, func=None, np_random=None):
        super().__init__(name, path, func, np_random)
        self.joint_range = None
        self.initial_joint = None
    
    def get_handle_pose(self, data):
        return get_pose(data, "handle", "geom")
    
    def get_jotinpos(self, data):
        return data.joint("task_joint").qpos[0]
    
    def get_perecenttage_opened(self, model, data):
        if self.joint_range is None:
            self.joint_range = model.jnt_range[mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, "task_joint")]
            self.initial_joint = self.get_jotinpos(data)
        current_task_joint = self.get_jotinpos(data)
        normalized_task_joint = abs(current_task_joint - self.initial_joint) / (
            abs(self.joint_range[1] - self.joint_range[0]) + 1e-8
        )
        return normalized_task_joint
    
    def open(self, model, data):
        if self.joint_range is None:
            self.joint_range = model.jnt_range[mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, "task_joint")]
            self.initial_joint = self.get_jotinpos(data)
        data.joint("task_joint").qpos = (
            self.joint_range[1] * self.np_random.uniform(0.5, 0.9)
            if self.joint_range[0] == 0
            else self.joint_range[0] * self.np_random.uniform(0.5, 0.9)
        )
        mujoco.mj_step(model, data, nstep=100)
        


import cv2
import torch
import torch.nn.functional as F
import numpy as np
from scipy.spatial.transform import Rotation as R
from egogym.policies.base_policy import BasePolicy
from baselines.rum.policy import Policy

class RUMPolicy(BasePolicy):
    
    def __init__(self, config=None, policy=None, **kwargs):
        super().__init__()

        if config is None:
            config = get_config()
        
        for key, value in kwargs.items():
            config[key] = value
            
        self.config = config
        self.name = "rum"
        self.device = config.get("device", "cuda" if torch.cuda.is_available() else "cpu")
        if policy is not None:
            self.policy = policy
        else:
            self.policy = Policy(model_path=config["model_path"], device=self.device)
        self.gripper_threshold = config["gripper_threshold"]
        self.grasping_style = config["grasping_style"]
        self.num_envs = None
        self.grasped = None
        self.desired_axis_t = torch.tensor([
            [-1,  0,  0],
            [ 0,  0, -1],
            [ 0, -1,  0]
        ], dtype=torch.float32, device=self.device)
        self.grasp_override_t = torch.tensor([0.00, 0.18, 0.04], dtype=torch.float32, device=self.device)

    def get_action(self, obs):

        if self.num_envs is None:
            if len(obs["rgb_ego"].shape) == 4:
                self.num_envs = obs["rgb_ego"].shape[0]
            else:
                self.num_envs = 1
            self.grasped = torch.zeros(self.num_envs, dtype=torch.bool, device=self.device)
        
        if self.num_envs == 1:
            obs = {key: obs[key][np.newaxis, ...] if isinstance(obs[key], np.ndarray) else obs[key] for key in obs}

        T_world_camera = torch.from_numpy(obs["camera_pose"]).view(self.num_envs, 4, 4).to(self.device)
        if "object_pose" in obs:
            T_world_object = torch.from_numpy(obs["object_pose"]).view(self.num_envs, 4, 4).to(self.device)
        else:
            T_world_object = torch.from_numpy(obs["handle_pose"]).view(self.num_envs, 4, 4).to(self.device)
        
        R_world_camera = T_world_camera[:, :3, :3]
        t_world_camera = T_world_camera[:, :3, 3]
        p_world = T_world_object[:, :3, 3]
        R_camera_world = R_world_camera.transpose(1, 2)
        p_camera = torch.bmm(R_camera_world, (p_world - t_world_camera).unsqueeze(-1)).squeeze(-1)
        object_3d_position = p_camera @ self.desired_axis_t.T
        object_3d_position[self.grasped] = self.grasp_override_t
        
        rgb_tensor = torch.from_numpy(obs["rgb_ego"]).permute(0, 3, 1, 2).to(self.device).float()
        rgb_resized = F.interpolate(rgb_tensor, size=(224, 224), mode='bilinear', align_corners=False)
        rgb_ego = rgb_resized.permute(0, 2, 3, 1).byte().cpu().numpy()

        rum_obs = {
            "rgb_ego": rgb_ego,
            "object_3d_position": object_3d_position.cpu().numpy(),
        }
        action_tensors = self.policy.infer(rum_obs)
        
        action_tensors_t = torch.from_numpy(action_tensors).to(self.device)
        R_action = R.from_euler("xyz", action_tensors[:, 3:6]).as_matrix()
        R_action_t = torch.from_numpy(R_action).to(self.device, dtype=torch.float32)
        
        action_pose = torch.eye(4, device=self.device).unsqueeze(0).repeat(self.num_envs, 1, 1)
        action_pose[:, :3, :3] = R_action_t
        action_pose[:, :3, 3] = action_tensors_t[:, :3]
        action_pose[:, :3, :3] = self.desired_axis_t.T @ action_pose[:, :3, :3] @ self.desired_axis_t
        action_pose[:, :3, 3] = action_pose[:, :3, 3] @ self.desired_axis_t
        
        self.grasped = torch.maximum(self.grasped, action_tensors_t[:, 6] < self.gripper_threshold)
        grasp_action = self.grasped.unsqueeze(1).float()
        if self.grasping_style == "continuous":
            grasp_action = 1 - action_tensors_t[:, 6].unsqueeze(1)

        actions = torch.cat([action_pose.view(self.num_envs, -1), grasp_action], dim=1)

        return actions.cpu().numpy()
    
    def reset(self, indicies=None):
        self.policy.reset(indicies)
        if indicies is not None and self.num_envs is not None:
            self.grasped[indicies] = False
        elif self.num_envs is not None:
            self.grasped = torch.zeros(self.num_envs, dtype=torch.bool, device=self.device)


def get_config():
    config = {
        "model_path": "checkpoints/rum_pick.pt",
        "gripper_threshold": 0.7,
        "grasping_style": "binary",
        "device": None,
    }
    return config



"""
A generic model wrapper for timm encoders.
"""

import pathlib
from typing import Iterable, Optional, Union
import warnings

import einops
import timm
import torch
import torch.nn as nn

from baselines.rum.models.encoders.abstract_base_encoder import AbstractEncoder
from baselines.rum.utils.decord_transforms import create_transform


class TimmModel(AbstractEncoder):
    def __init__(
        self,
        model_name: str = "hf-hub:notmahi/dobb-e",
        pretrained: bool = True,
        weight_path: Union[None, str, pathlib.Path] = None,
    ):
        super().__init__()
        self._model_name = model_name

        self.model = timm.create_model(model_name, pretrained=pretrained, num_classes=0)
        if weight_path:
            self.load_weights(weight_path, strict=False)

    def transform(self, x):
        return x

    @property
    def feature_dim(self):
        return self.model.num_features

    def to(self, device):
        self.model.to(device)
        return self

    def forward(self, x):
        return self.model(self.transform(x))


class TimmSSL(TimmModel):
    def __init__(
        self,
        model_name: str = "hf-hub:notmahi/dobb-e",
        pretrained=True,
        override_aug_kwargs=dict(
            hflip=0.0, vflip=0.0, scale=(1.0, 1.0), crop_pct=0.875
        ),
        weight_path: Union[None, str, pathlib.Path] = None,
    ):
        super().__init__(model_name, pretrained=pretrained)
        data_cfg = timm.data.resolve_data_config(self.model.pretrained_cfg)
        # Now define the transforms.
        data_cfg.update(override_aug_kwargs)
        data_cfg["is_training"] = True
        self._train_transform = create_transform(**data_cfg)
        data_cfg["is_training"] = False
        self._test_transform = create_transform(**data_cfg)
        if weight_path is not None:
            self.load_weights(weight_path, strict=True)

    def transform(self, x):
        return (
            self._train_transform(x) if self.model.training else self._test_transform(x)
        )

    def forward(self, x):
        # Split the input into frames and labels.
        images, *labels = x
        # Flatten the frames into a single batch.
        flattened_images = einops.rearrange(images, "b t c h w -> (b t) c h w")
        # Transform and pass through the model.
        result = self.model(self.transform(flattened_images))
        # Unflatten the result.
        result = einops.rearrange(result, "(b t) c -> b t c", b=images.shape[0])
        return result



from abc import ABC, abstractmethod
from typing import Optional

import torch

class BinarizeGripper(ABC):
    def __init__(
        self,
        model: Optional[torch.nn.Module],
        binarize_gripper: bool = False,
        threshold: float = 0.5,
        upper_value: float = 1.0,
        lower_value: float = 0.0,
        *args,
        **kwargs,
    ):
        super().__init__()
        assert (
            not binarize_gripper or not model.relative_gripper
        ), "Binarize gripper and relative gripper cannot be used together"

        self.binarize_gripper = binarize_gripper
        self.threshold = threshold
        self.upper_value = upper_value
        self.lower_value = lower_value

    @abstractmethod
    def binarize_gripper(self, actions):
        # here the actions will be of shape (batch_size, seq_len, action_space)
        # last element in action space is gripper values between 0 and 1

        # binarize the gripper values
        if self.binarize_gripper:
            actions[:, :, -1] = torch.where(
                actions[:, :, -1] > self.threshold, self.upper_value, self.lower_value
            )
        return actions



import abc

import torch


class AbstractLossFn(torch.nn.Module, abc.ABC):
    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        pass

    @abc.abstractmethod
    def forward(self, data, output, *args, **kwargs):
        raise NotImplementedError



import os
import numpy as np
import time
import re
import io
import json
import base64
from PIL import Image
from egogym.utils import pixel_to_world

MODEL = "gpt-4o"
PROMPT = """Get all points matching the following object: {object_name}. The label returned should be an identifying name for the object detected.
The answer should follow the json format: [{{"point": [y, x], "label": "{object_name}"}}, ...]. The points are in [y, x] format normalized to 0-1000."""

class UnprivilegedChatGPT:
    
    def __init__(self, env, api_key=None):
        try:
            from openai import OpenAI
        except ImportError:
            raise ImportError("openai package is not installed. Please install it with `pip install openai`.")

        self.env = env
        self.unprivileged_T_world_object = None
        
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")
    
        
        self.client = OpenAI(api_key=api_key)
    
    def _call_api_with_retry(self, messages, max_retries=3, base_delay=2):
        """Call OpenAI API with retry logic for transient errors."""
        for attempt in range(max_retries):
            try:
                return self.client.chat.completions.create(
                    model=MODEL,
                    messages=messages,
                    temperature=0.5,
                    max_tokens=1024
                )
            except Exception as e:
                error_code = getattr(e, 'status_code', None)
                if error_code in [503, 429] and attempt < max_retries - 1:
                    delay = base_delay * (2 ** attempt)
                    print(f"API error {error_code}: {str(e)}. Retrying in {delay}s... (attempt {attempt + 1}/{max_retries})")
                    time.sleep(delay)
                    continue
                else:
                    raise
    
    def _parse_json_point_yx(self, text):
        """Parse JSON format with point coordinates in [y, x] format normalized to 0-1000."""
        try:
            json_match = re.search(r'```(?:json)?\s*(\[.*?\])\s*```', text, re.DOTALL)
            if json_match:
                json_str = json_match.group(1)
            else:
                json_match = re.search(r'(\[.*?\])', text, re.DOTALL)
                if json_match:
                    json_str = json_match.group(1)
                else:
                    return None
            
            data = json.loads(json_str)
            if not isinstance(data, list) or len(data) == 0:
                return None
            
            point_data = data[0]
            if 'point' in point_data:
                point = point_data['point']
                if isinstance(point, list) and len(point) >= 2:
                    return (float(point[0]), float(point[1]))
            
            return None
            
        except (json.JSONDecodeError, KeyError, ValueError, IndexError) as e:
            return None
    
    def _infer_point_sync(self, rgb: np.ndarray, object_name: str) -> np.ndarray:
        """Use OpenAI GPT-4o API to locate object in image and return normalized coordinates."""
        if rgb.dtype == np.float32 or rgb.dtype == np.float64:
            rgb = (rgb * 255).astype(np.uint8)
        image = Image.fromarray(rgb)
        img_width, img_height = image.size
        
        prompt = PROMPT.format(object_name=object_name)
        
        try:
            buf = io.BytesIO()
            image.save(buf, format='PNG')
            image_bytes = buf.getvalue()
            base64_image = base64.b64encode(image_bytes).decode('utf-8')
            
            messages = [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/png;base64,{base64_image}"
                            }
                        },
                        {
                            "type": "text",
                            "text": prompt
                        }
                    ]
                }
            ]
            
            response = self._call_api_with_retry(messages)
            generated_text = response.choices[0].message.content
            
            json_point = self._parse_json_point_yx(generated_text)
            if json_point:
                y_pos, x_pos = json_point
                x_norm = float(x_pos) / 1000.0
                y_norm = float(y_pos) / 1000.0
                
                x_norm = max(0.0, min(1.0, x_norm))
                y_norm = max(0.0, min(1.0, y_norm))
                
                return np.array([x_norm, y_norm], dtype=np.float32)
            
            return np.array([0.5, 0.5], dtype=np.float32)
            
        except Exception as e:
            return np.array([0.5, 0.5], dtype=np.float32)

    def reset(self, **kwargs):
        obs, info = self.env.reset(**kwargs)
        
        robot = self.env.unwrapped.robot
        robot.rgb_renderers[robot.camera_names[0]].enable_depth_rendering()
        robot.rgb_renderers[robot.camera_names[0]].update_scene(robot.data, robot.camera_names[0])
        depth = robot.rgb_renderers[robot.camera_names[0]].render()
        robot.rgb_renderers[robot.camera_names[0]].disable_depth_rendering()

        point_norm = self._infer_point_sync(obs["rgb_ego"], obs["object_name"])
        x_norm, y_norm = point_norm
        x = int(x_norm * self.env.unwrapped.render_width)
        y = int(y_norm * self.env.unwrapped.render_height)
        depth_value = depth[y, x] + 0.03
        self.unprivileged_T_world_object = pixel_to_world(x, y, depth_value, self.env.unwrapped.model, self.env.unwrapped.data, robot.camera_names[0], self.env.unwrapped.render_width, self.env.unwrapped.render_height)

        return obs, info
    
    def step(self, action):
        obs, reward, terminated, truncated, info = self.env.step(action)
        
        object_pose = obs["object_pose"].reshape(4,4)
        object_pose[:3, 3] = self.unprivileged_T_world_object
        obs["object_pose"] = object_pose.flatten()

        return obs, reward, terminated, truncated, info
    
    def close(self):
        self.env.close()
    
    def __getattr__(self, name):
        return getattr(self.env, name)



from egogym.embodiments.arms.arm import Arm
from egogym.embodiments.robot import requires_mj
from egogym.utils import get_pose
import numpy as np


class DroidArm(Arm):    
    def __init__(self, control_steps=600, grasping_steps=100, action_space="delta", np_random=None):
        super().__init__(control_steps=control_steps, grasping_steps=grasping_steps, action_space=action_space, np_random=np_random)
        self.name = "droid"
        self.initial_q = np.array([0, -1/5 * np.pi, 0, -4/5 * np.pi, 0, 3/5 * np.pi, 0.0])
        self.joint_names = ["joint1", "joint2", "joint3", "joint4", "joint5", "joint6", "joint7"]
        self.gripper_actuator_name = "fingers_actuator"
    
    @requires_mj
    def prepare(self, desired_pose, step_fn=None):

        desired_pose[:3,3] += np.array([0.0, 0.6, 0.075])
        if self.add_noise_init:
            position_noise = self.np_random.uniform(-0.03, 0.03, size=3)
            desired_pose[:3, 3] += position_noise

        self.data.ctrl[self.act_ids] = self.initial_q
        step_fn(500)
        gripper_position = self.get_tcp_pose()[:3, 3]
        link_position = get_pose(self.data, "link0", "body")[:3, 3]
        delta_position = gripper_position - link_position
        desired_gripper_position = desired_pose[:3, 3]
        desired_link_position = desired_gripper_position - delta_position
        self.model.body_pos[self.model.body("link0").id] = desired_link_position
        step_fn(100)

    @requires_mj
    def get_grasp(self):
        val = self.data.qpos[self.model.joint("right_driver_joint").qposadr[0]] / 0.8102
        if val < 0.01:
            val = 0.0
        val = np.array([val], dtype=np.float32).reshape(1)
        current = np.array([(np.clip(val, 0.0, 1.0))], dtype=np.float32).reshape(1)[0]
        return current
    
    @requires_mj
    def set_grasp(self, grip_action):
        desired_grasp = 255.0 * grip_action
        self.data.ctrl[self.model.actuator("fingers_actuator").id] = desired_grasp.item()
        if desired_grasp != self.last_grasp:
            self.last_grasp = desired_grasp
            return 1
        return 0


import logging
from enum import Enum
from typing import Dict, Optional, Tuple

import einops
import torch
import torch.nn as nn
import torch.nn.functional as F
import tqdm

from baselines.rum.models.bet.gpt import GPT
from baselines.rum.models.bet.utils import MLP
from baselines.rum.models.bet.vqvae.vqvae import VqVae

GENERATOR_SEED_FIXED = 123456789


class VQBehaviorTransformer(nn.Module):
    GOAL_SPEC = Enum("GOAL_SPEC", "concat stack unconditional")

    def __init__(
        self,
        obs_dim: int,
        act_dim: int,
        goal_dim: int,
        gpt_model: GPT,
        vqvae_model: VqVae,
        offset_loss_multiplier: float = 1.0e2,
        secondary_code_multiplier: float = 0.5,
        gamma: float = 2.0,
        obs_window_size=10,
        act_window_size=10,
        sequentially_select=False,
        use_og_bet_loss=False,
        use_half_and_half_loss=False,
        temperature=1.0,
        device="cuda",
    ):
        super().__init__()
        self._obs_dim = obs_dim
        self._act_dim = act_dim
        self._goal_dim = goal_dim
        self.obs_window_size = obs_window_size
        self.act_window_size = act_window_size
        self.sequentially_select = sequentially_select
        self._use_og_bet_loss = use_og_bet_loss
        self._use_half_and_half_loss = use_half_and_half_loss
        self.temperature = temperature
        self.device = device

        if goal_dim <= 0:
            self._cbet_method = self.GOAL_SPEC.unconditional
        # elif obs_dim == goal_dim:
        #     self._cbet_method = self.GOAL_SPEC.concat
        # TODO (haritheja): this is a temporary fix, we should be able to handle different types of goals like concat or stack
        else:
            self._cbet_method = self.GOAL_SPEC.stack

        self._gpt_model = gpt_model
        self._vqvae_model = vqvae_model
        self._G = self._vqvae_model.vqvae_groups  # G(number of groups)
        self._C = self._vqvae_model.vqvae_n_embed  # C(number of code integers)
        self._D = self._vqvae_model.embedding_dim  # D(embedding dims)
        self._current_steps = 0
        if self.sequentially_select:
            self._map_to_cbet_preds_bin1 = MLP(
                in_channels=gpt_model.config.output_dim,
                hidden_channels=[512, 512, self._C],
            )
            self._map_to_cbet_preds_bin2 = MLP(
                in_channels=gpt_model.config.output_dim + self._C,
                hidden_channels=[512, self._C],
            )
        else:
            self._map_to_cbet_preds_bin = MLP(
                in_channels=gpt_model.config.output_dim,
                hidden_channels=[1024, 1024, self._G * self._C],
            )
        self._map_to_cbet_preds_offset = MLP(
            in_channels=gpt_model.config.output_dim,
            hidden_channels=[
                1024,
                1024,
                self._G * self._C * (act_dim * self.act_window_size),
            ],
        )
        self._offset_loss_multiplier = offset_loss_multiplier
        self._secondary_code_multiplier = secondary_code_multiplier
        self._criterion = FocalLoss(gamma=gamma)

    def forward(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: Optional[torch.Tensor],
        second_half: bool = False,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
        # VQ-BeT doesn't use "padding_seq" and "predict_with_offset" input
        return self._predict(obs_seq, goal_seq, action_seq, second_half)

    def _predict(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: Optional[torch.Tensor],
        second_half: bool,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Dict[str, float]]:
        """
        Assume dimensions are N T D for N sequences of T timesteps with dimension D.

        obs_seq : (Batch size) X (obs window) X (obs dim)
        goal_seq : (Batch size) X (goal window) X (obs dim)
        action_seq : (Batch size) X (total window) X (action dim)

        assume that
        total_w (total window) of action sequence is
        total_w = obs_w + act_w - 1
        e.g. is observation window is 5, and antion pred window is 3,
        the obs seq (o_{t-4}, o_{t-3}, o_{t-2}, o_{t-1}, o_{t}) will predict (a_{t}, a_{t+1}, a_{t+2})

        However, we not only predict (a_{t}, a_{t+1}, a_{t+2}), but also all the action seq for all the actions in the sequence

        o_{t-4} -> |   | -> (a_{t-4}, a_{t-3}, a_{t-2})
        o_{t-3} -> | G | -> (a_{t-3}, a_{t-2}, a_{t-1})
        o_{t-2} -> | P | -> (a_{t-2}, a_{t-1}, a_{t})
        o_{t-1} -> | T | -> (a_{t-1}, a_{t}, a_{t+1})
        o_{t}   -> |   | -> (a_{t}, a_{t+1}, a_{t+2})
        """
        if obs_seq.shape[1] < self.obs_window_size:
            # if input size is smaller than obs_window size (e.g. the initial steps of env eval episodes,
            # VQ-BeT copy the obs and tile it to match obs_window_size
            obs_seq = torch.cat(
                (
                    torch.tile(
                        obs_seq[:, 0, :],
                        (1, self.obs_window_size - obs_seq.shape[1], 1),
                    ),
                    obs_seq,
                ),
                dim=-2,
            )
            if self._cbet_method == self.GOAL_SPEC.stack:
                goal_seq = torch.cat(
                    (
                        torch.tile(
                            goal_seq[:, 0, :],
                            (1, self.obs_window_size - goal_seq.shape[1], 1),
                        ),
                        goal_seq,
                    ),
                    dim=-2,
                )
        if self._cbet_method == self.GOAL_SPEC.unconditional:
            gpt_input = obs_seq
        elif self._cbet_method == self.GOAL_SPEC.concat:
            gpt_input = torch.cat([goal_seq, obs_seq], dim=1)
        elif self._cbet_method == self.GOAL_SPEC.stack:
            gpt_input = torch.cat([goal_seq, obs_seq], dim=-1)
        else:
            raise NotImplementedError

        gpt_output = self._gpt_model(gpt_input)
        if self._cbet_method == self.GOAL_SPEC.concat:
            # Chop off the goal encodings.
            gpt_output = gpt_output[:, goal_seq.size(1) :, :]
        gpt_output = einops.rearrange(gpt_output, "N T (G C) -> (N T) (G C)", G=self._G)
        obs = einops.rearrange(obs_seq, "N T O -> (N T) O")
        obs = obs.unsqueeze(dim=1)
        # note that output of offset network is G C WA,
        # where G is number of 'layers' of Residual VQ-VAE
        # C is number of words in each layer's dictionary
        # and W, A is predicted action window, and predicted action dims
        if self.sequentially_select:
            cbet_logits1 = self._map_to_cbet_preds_bin1(gpt_output)
            cbet_offsets = self._map_to_cbet_preds_offset(gpt_output)
            cbet_offsets = einops.rearrange(
                cbet_offsets, "(NT) (G C WA) -> (NT) G C WA", G=self._G, C=self._C
            )
            cbet_probs1 = torch.softmax(cbet_logits1 / self.temperature, dim=-1)
            NT, choices = cbet_probs1.shape
            G = self._G
            sampled_centers1 = einops.rearrange(
                torch.multinomial(cbet_probs1.view(-1, choices), num_samples=1),
                "(NT) 1 -> NT",
                NT=NT,
            )
            cbet_logits2 = self._map_to_cbet_preds_bin2(
                torch.cat(
                    (gpt_output, F.one_hot(sampled_centers1, num_classes=self._C)),
                    axis=1,
                )
            )
            cbet_probs2 = torch.softmax(cbet_logits2 / self.temperature, dim=-1)
            sampled_centers2 = einops.rearrange(
                torch.multinomial(cbet_probs2.view(-1, choices), num_samples=1),
                "(NT) 1 -> NT",
                NT=NT,
            )
            sampled_centers = torch.stack(
                (sampled_centers1, sampled_centers2), axis=1
            )  # NT, G
        else:
            cbet_logits = self._map_to_cbet_preds_bin(gpt_output)
            cbet_offsets = self._map_to_cbet_preds_offset(gpt_output)
            cbet_logits = einops.rearrange(
                cbet_logits, "(NT) (G C) -> (NT) G C", G=self._G
            )
            cbet_offsets = einops.rearrange(
                cbet_offsets, "(NT) (G C WA) -> (NT) G C WA", G=self._G, C=self._C
            )
            cbet_probs = torch.softmax(cbet_logits / self.temperature, dim=-1)
            NT, G, choices = cbet_probs.shape
            sampled_centers = einops.rearrange(
                torch.multinomial(cbet_probs.view(-1, choices), num_samples=1),
                "(NT G) 1 -> NT G",
                NT=NT,
            )
        if action_seq is not None:
            n, total_w, act_dim = action_seq.shape
            act_w = self._vqvae_model.input_dim_h
            obs_w = total_w + 1 - act_w
            output_shape = (n, obs_w, act_w, act_dim)
            output = torch.empty(output_shape).to(action_seq.device)
            for i in range(obs_w):
                output[:, i, :, :] = action_seq[:, i : i + act_w, :]
            action_seq = einops.rearrange(output, "N T W A -> (N T) W A")
            _, action_bins = self._vqvae_model.get_code(
                action_seq, obs
            )  # action_bins: NT, G

        with torch.no_grad():
            centers = self._vqvae_model.draw_code_forward(sampled_centers).view(
                NT, -1, self._D
            )
            return_decoder_input = einops.rearrange(
                centers.clone().detach(), "NT G D -> NT (G D)"
            )
            decoded_action = (
                self._vqvae_model.get_action_from_latent(return_decoder_input, obs)
                .clone()
                .detach()
            )  # NT, A

        def get_offset_from_centers(centers):
            indices = (
                torch.arange(NT, device=self.device).unsqueeze(1),
                torch.arange(self._G, device=self.device).unsqueeze(0),
                centers,
            )
            # Use advanced indexing to sample the values
            sampled_offsets = cbet_offsets[indices]  # NT, G, WA or NT, G, A
            sampled_offsets = sampled_offsets.sum(dim=1)
            sampled_offsets = einops.rearrange(
                sampled_offsets, "NT (W A) -> NT W A", W=self._vqvae_model.input_dim_h
            )
            return sampled_offsets

        if self._use_og_bet_loss and action_seq is not None:
            sampled_offsets = get_offset_from_centers(action_bins)
        else:
            sampled_offsets = get_offset_from_centers(sampled_centers)

        a_hat = decoded_action + sampled_offsets

        if action_seq is None:
            return a_hat, None, {}
        # Figure out the loss for the actions.
        # First, we need to find the GT VQ codes for each action.

        # Now we can compute the loss.
        if action_seq.ndim == 2:
            action_seq = action_seq.unsqueeze(0)

        offset_target = action_seq - decoded_action
        if self._use_half_and_half_loss and action_seq is not None:
            offset_gt = get_offset_from_centers(action_bins)
            offset_sampled = get_offset_from_centers(sampled_centers)
            offset_loss = 0.5 * torch.nn.L1Loss()(
                offset_target, offset_gt
            ) + 0.5 * torch.nn.L1Loss()(offset_target, offset_sampled)
        else:
            offset_loss = torch.nn.L1Loss()(offset_target, sampled_offsets)

        if self.sequentially_select:
            cbet_loss1 = self._criterion(  # F.cross_entropy
                cbet_logits1[:, :],
                action_bins[:, 0],
            )
            cbet_logits2 = self._map_to_cbet_preds_bin2(
                torch.cat(
                    (gpt_output, F.one_hot(action_bins[:, 0], num_classes=self._C)),
                    axis=1,
                )
            )
            cbet_loss2 = self._criterion(  # F.cross_entropy
                cbet_logits2[:, :],
                action_bins[:, 1],
            )
        else:
            cbet_loss1 = self._criterion(  # F.cross_entropy
                cbet_logits[:, 0, :],
                action_bins[:, 0],
            )
            cbet_loss2 = self._criterion(  # F.cross_entropy
                cbet_logits[:, 1, :],
                action_bins[:, 1],
            )
        cbet_loss = cbet_loss1 * 5 + cbet_loss2 * self._secondary_code_multiplier

        equal_total_code_rate = (
            torch.sum(
                (torch.sum((action_bins == sampled_centers).int(), axis=1) == G).int()
            )
            / NT
        )
        equal_primary_code_rate = torch.sum(
            (action_bins[:, 0] == sampled_centers[:, 0]).int()
        ) / (NT)
        equal_secondary_code_rate = torch.sum(
            (action_bins[:, 1] == sampled_centers[:, 1]).int()
        ) / (NT)
        # if second_half:
        #     cbet_loss = cbet_loss * 0
        loss = cbet_loss + self._offset_loss_multiplier * offset_loss
        action_mse = F.mse_loss(a_hat, action_seq, reduction="none")
        action_l1 = F.l1_loss(a_hat, action_seq, reduction="none")
        norm = torch.norm(action_seq, p=2, dim=-1, keepdim=True) + 1e-9
        normalized_mse = (action_mse / norm).mean()

        translation_loss = F.mse_loss(a_hat[:, :, :3], action_seq[:, :, :3]).detach()
        rotation_loss = F.mse_loss(a_hat[:, :, 3:6], action_seq[:, :, 3:6]).detach()
        gripper_loss = F.mse_loss(a_hat[:, :, 6:], action_seq[:, :, 6:]).detach()

        loss_dict = {
            "classification_loss": cbet_loss.detach().cpu().item(),
            "offset_loss": offset_loss.detach().cpu().item(),
            "loss": loss.detach().cpu().item(),
            "equal_total_code_rate": equal_total_code_rate,
            "equal_primary_code_rate": equal_primary_code_rate,
            "equal_secondary_code_rate": equal_secondary_code_rate,
            "L2_loss": action_mse.mean().detach().cpu().item(),
            "L2_loss_normalized": normalized_mse.mean().detach().cpu().item(),
            "L1_loss": action_l1.mean().detach().cpu().item(),
            "translation_loss": translation_loss,
            "rotation_loss": rotation_loss,
            "gripper_loss": gripper_loss,
        }
        return a_hat, loss, loss_dict

    # def configure_optimizers(self, weight_decay, learning_rate, betas):

    #     optimizer1 = self._gpt_model.configure_optimizers(
    #         weight_decay=weight_decay,
    #         learning_rate=learning_rate,
    #         betas=betas,
    #     )
    #     if self.sequentially_select:
    #         optimizer1.add_param_group({"params": self._map_to_cbet_preds_bin1.parameters()})
    #         optimizer1.add_param_group({"params": self._map_to_cbet_preds_bin2.parameters()})
    #     else:
    #         optimizer1.add_param_group({"params": self._map_to_cbet_preds_bin.parameters()})
    #     optimizer2 = torch.optim.AdamW(self._map_to_cbet_preds_offset.parameters(), lr=learning_rate, weight_decay=weight_decay, betas=betas)
    #     return {"optimizer1": optimizer1, "optimizer2": optimizer2}

    def _begin_epoch(self, optimizer, **kwargs):
        # log codebook usage rate for debugging
        # lr_0 = optimizer.param_groups[0]["lr"]
        # lr_neg1 = optimizer.param_groups[-1]["lr"]
        # return {"lr_0": lr_0, "lr_neg1": lr_neg1}
        return None

    def _load_from_state_dict(self, *args, **kwargs):
        # Don't fit kmeans if we are loading from a state dict.
        # if (path / "cbet_model.pt").exists():
        #     self.load_state_dict(torch.load(path / "cbet_model.pt"))
        # elif (path / "gpt_model.pt").exists():
        #     self._gpt_model.load_state_dict(torch.load(path / "gpt_model.pt"))
        # else:
        #     logging.warning("No model found at %s", path)
        return super()._load_from_state_dict(*args, **kwargs)

    # def load_model(self, path):
    #     if (path / "cbet_model.pt").exists():
    #         self.load_state_dict(torch.load(path / "cbet_model.pt"))
    #     elif (path / "gpt_model.pt").exists():
    #         self._gpt_model.load_state_dict(torch.load(path / "gpt_model.pt"))
    #     else:
    #         logging.warning("No model found at %s", path)


class FocalLoss(nn.Module):
    def __init__(self, gamma: float = 0, reduction: str = "mean"):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        if reduction not in ("mean", "sum", "none"):
            raise NotImplementedError
        self.reduction = reduction

    def forward(self, input, target):
        logpt = F.log_softmax(input, dim=-1)
        logpt = logpt.gather(1, target.view(-1, 1)).view(-1)
        pt = logpt.exp()

        loss = -1 * (1 - pt) ** self.gamma * logpt
        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        else:
            return loss



import math
import numpy as np
import torch
import hydra
from scipy.spatial.transform import Rotation as R
import os
import xml.etree.ElementTree as ET
import gdown
import zipfile
import random
from PIL import Image
import importlib.util
import sys
from pathlib import Path


def load_policy_from_file(policy_path, config_overrides=None):
    policy_path = Path(policy_path)
    
    spec = importlib.util.spec_from_file_location("policy_module", policy_path)
    policy_module = importlib.util.module_from_spec(spec)
    sys.modules["policy_module"] = policy_module
    spec.loader.exec_module(policy_module)
    
    config = policy_module.get_config()
    
    if config_overrides:
        for key, value in config_overrides.items():
            config[key] = value
    
    policy_class = None
    for attr_name in dir(policy_module):
        attr = getattr(policy_module, attr_name)
        if isinstance(attr, type) and attr_name.endswith('Policy') and attr_name != 'BasePolicy':
            if hasattr(attr, '__module__') and attr.__module__ == 'policy_module':
                policy_class = attr
                break
    
    if policy_class is None:
        raise ValueError(f"No policy class found in {policy_path}")
    
    return policy_class(config=config)


def maybe_download():
    objects_path = os.path.join(os.path.dirname(__file__), "assets/objects")
    if not os.path.exists(objects_path):
        print("Downloading Objects...")
        download_and_extract_zip(
            gdrive_link="https://drive.google.com/file/d/1wDczBM6H3rQvFUGKe_vQINE5Bm2fBlU4",
            output_dir=os.path.join(os.path.dirname(__file__), "assets"),
        )
    return objects_path

def get_textures(root_folder=None):
    results = []
    if root_folder is None:
        root_folder = os.path.join(os.path.dirname(__file__), "assets/textures")

    for dirpath, _, filenames in os.walk(root_folder):
        for filename in filenames:
            if filename.endswith('.png'):
                full_path = os.path.join(dirpath, filename)
                rel_path = os.path.relpath(full_path, root_folder)
                results.append(rel_path)
    
    return results

def get_objects(root_folder=None):
    results = []
    if root_folder is None:
        root_folder = maybe_download()

    for dirpath, _, filenames in os.walk(root_folder):
        if "model.xml" in filenames:
            full_path = os.path.join(dirpath, "model.xml")
            parent_folder = os.path.basename(dirpath)
            results.append((parent_folder, full_path))

    results.append(("cabinet", "procedurally_generated"))
    results.append(("drawer", "procedurally_generated"))
    
    return results

def make_objects_manager(objects_set, np_random, shuffle=True):
    from egogym.managers import ObjectsManager
    objects = []
    available_objects = get_objects()
    available_objects = {obj[0]: obj[1] for obj in available_objects}
    for obj_name in objects_set:
        if obj_name in available_objects:
            objects.append((obj_name, available_objects[obj_name]))
    return ObjectsManager(objects, np_random, shuffle=shuffle)

def make_textures_manager(textures_set, np_random):
    from egogym.managers import TexturesManager
    textures = []
    available_textures = get_textures(root_folder=os.path.join(os.path.dirname(__file__), "assets/textures"))
    for tex_path in textures_set:
        for available_tex_name in available_textures:
            if tex_path in available_tex_name:
                textures.append(tex_path)
    return TexturesManager(textures, np_random)

def include_in_scene(scene_xml_content, include_file_path):
    root = ET.fromstring(scene_xml_content)
    new_include = ET.Element("include", file=include_file_path)
    includes = root.findall("include")
    if includes:
        index = list(root).index(includes[-1]) + 1
        root.insert(index, new_include)
    else:
        root.insert(0, new_include)
    return ET.tostring(root, encoding="unicode")

def download_and_extract_zip(gdrive_link, output_dir):
    os.makedirs(output_dir, exist_ok=True)
    file_id = gdrive_link.split("/d/")[1].split("/")[0]
    download_url = f"https://drive.google.com/uc?id={file_id}"
    zip_path = os.path.join(output_dir, "temp_download.zip")

    gdown.download(download_url, zip_path, quiet=False)

    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        zip_ref.extractall(output_dir)

    os.remove(zip_path)

def euler_to_mat_xyz(rx, ry, rz):
    cx = torch.cos(rx)
    sx = torch.sin(rx)
    cy = torch.cos(ry)
    sy = torch.sin(ry)
    cz = torch.cos(rz)
    sz = torch.sin(rz)
    B = rx.shape[0]
    Rmats = torch.empty(B, 3, 3, device=rx.device)
    Rmats[:, 0, 0] = cy * cz
    Rmats[:, 0, 1] = -cy * sz
    Rmats[:, 0, 2] = sy
    Rmats[:, 1, 0] = cx * sz + cz * sx * sy
    Rmats[:, 1, 1] = cx * cz - sx * sy * sz
    Rmats[:, 1, 2] = -cy * sx
    Rmats[:, 2, 0] = sx * sz - cx * cz * sy
    Rmats[:, 2, 1] = cz * sx + cx * sy * sz
    Rmats[:, 2, 2] = cx * cy
    return Rmats

def make_pose(pos, rot):
    pose = np.eye(4)
    pose[:3, 3] = pos
    pose[:3, :3] = rot.as_matrix()
    return pose

def get_pos_rot(data, name, type):
    type_lower = type.lower()
    if type_lower == "body-control":
        pos = data.body(name).xpos.copy()
        rot_mat = data.body(name).xmat.copy().reshape((3, 3))
        return pos, rot_mat
    if type_lower == "body":
        pos = data.body(name).xipos.copy()
        rot_mat = data.body(name).xmat.copy().reshape((3, 3))
        return pos, rot_mat
    elif type_lower == "site":
        pos = data.site(name).xpos.copy()
        rot_mat = data.site(name).xmat.copy().reshape((3, 3))
        return pos, rot_mat
    elif type_lower == "geom":
        pos = data.geom(name).xpos.copy()
        rot_mat = data.geom(name).xmat.copy().reshape((3, 3))
        return pos, rot_mat
    elif type_lower == "camera":
        cam_id = data.model.camera(name).id
        pos = data.cam_xpos[cam_id].copy()
        rot_mat = data.cam_xmat[cam_id].copy().reshape((3, 3))
        return pos, rot_mat
    else:
        raise ValueError(
            f'type_ must be one of "geom", "body", or "site", got "{type}".'
        )

def get_pose(data, name, type):
    pos, rot_mat = get_pos_rot(data, name, type)
    pose = np.eye(4)
    pose[:3, 3] = pos
    pose[:3, :3] = rot_mat
    return pose

def position_sampler(
    np_random,
    num_samples,
    box_half_ranges,
    min_distance,
    max_total_tries=100,
    start_position=None,
    thickness=0.1,
):
    hx, hy = map(float, box_half_ranges)
    samples = []
    samples_arr = []
    cell = float(min_distance) / np.sqrt(2.0)
    ox, oy = -hx, -hy
    grid = {}

    def to_cell_idx(p):
        return (
            int(np.floor((p[0] - ox) / cell)),
            int(np.floor((p[1] - oy) / cell)),
        )

    def is_far_enough(p):
        cx, cy = to_cell_idx(p)
        for dx in (-1, 0, 1):
            for dy in (-1, 0, 1):
                key = (cx + dx, cy + dy)
                j = grid.get(key)
                if j is None:
                    continue
                q = samples_arr[j]
                if (p[0] - q[0]) ** 2 + (p[1] - q[1]) ** 2 < min_distance**2:
                    return False
        return True

    tries = 0
    while len(samples) < num_samples and tries < max_total_tries:
        tries += 1
        px = np_random.uniform(-hx, hx)
        py = np_random.uniform(-hy, hy)
        p = (px, py)
        if not samples_arr:
            samples.append([px, py])
            samples_arr.append(np.array(p, dtype=float))
            grid[to_cell_idx(p)] = 0
            continue
        if is_far_enough(p):
            idx = len(samples)
            samples.append([px, py])
            samples_arr.append(np.array(p, dtype=float))
            grid[to_cell_idx(p)] = idx

    if start_position is None:
        return samples

    A = np.array(start_position, dtype=float)
    blocked = [False] * len(samples)
    eps = 1e-9

    S = np.array(samples, dtype=float)
    for i in range(len(samples)):
        B = S[i]
        AB = B - A
        ab2 = float(AB[0] * AB[0] + AB[1] * AB[1])
        if ab2 <= eps:
            blocked[i] = False
            continue
        for j in range(len(samples)):
            if j == i:
                continue
            P = S[j]
            t = float(np.dot(P - A, AB) / ab2)
            if t <= eps or t >= 1.0 - eps:
                continue
            C = A + t * AB
            if float(np.sum((P - C) ** 2)) <= thickness * thickness:
                blocked[i] = True
                break

    if blocked[0]:
        for i in range(1, len(blocked)):
            if not blocked[i]:
                samples[0], samples[i] = samples[i], samples[0]
                blocked[0], blocked[i] = blocked[i], blocked[0]
                break
    return samples


def add_xml_to_scene(original_xml_string, new_xml_string):
    """
    Merges new_xml_string into original_xml_string by adding elements from corresponding sections.
    Elements from new XML's worldbody go under original's worldbody, assets under assets, etc.
    """

    original_root = ET.fromstring(original_xml_string)
    new_root = ET.fromstring(new_xml_string)

    for new_section in new_root:
        section_name = new_section.tag

        original_section = original_root.find(section_name)

        if original_section is not None:
            for child in new_section:
                original_section.append(child)
        else:
            original_root.append(new_section)

    return ET.tostring(original_root, encoding="unicode")

def make_handle_xml(
    handle_type="pullbar",
    handle_name="handle",
    position=(0, 0, 0),
    euler_orientation=(0, 0, 0),
    handle_radius=0.009,
    handle_length=0.05,
    handle_material="M_hinge_metal",
    mass=0.1,
    friction_geom="0 0 0",
    hinge_side="left",
    np_random=None,
):
    pos_x, pos_y, pos_z = position
    euler_x, euler_y, euler_z = euler_orientation

    if handle_type == "knob":
        if handle_name == "handle":
            knob_geom_name = "handle"
            knob_geom_viz_name = "handle_viz"
        else:
            knob_geom_name = f"knob_geo_{hinge_side}"
            knob_geom_viz_name = f"knob_geom_viz_{hinge_side}"
        friction_geom = "5 5 5"
        handle_body_xml = f"""
        <body name="{handle_name}_body" pos="{pos_x + (-0.02 if hinge_side == "left" else 0.02)} {pos_y + 0.07} {pos_z}" euler="{euler_x} {euler_y} 0">
                    <body name="knob">
                    <geom name="{knob_geom_name}" mesh="knob" class="visual" material="{handle_material}"/>
                    <geom name="{knob_geom_viz_name}" mesh="knob_collision_0" rgba="0.12319198224928918 0.8185897489755214 0.9473380251850289 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_1" rgba="0.9952048472348035 0.9587846043226812 0.44813684566262046 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_2" rgba="0.5562189501844684 0.3964830683252689 0.24776822693559275 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_3" rgba="0.5665588588501549 0.4670123537790013 0.1620234740959262 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_4" rgba="0.6170081600309119 0.6175499636469116 0.07870231532672867 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_5" rgba="0.5008884031488108 0.5029776170199916 0.7335611007545255 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_6" rgba="0.6962939670098736 0.9990710674845018 0.1259585452380786 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_7" rgba="0.25804299152558585 0.3180969182215061 0.9745095357400368 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_8" rgba="0.012626418834853048 0.6472133449453128 0.6250494987727182 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    <geom mesh="knob_collision_9" rgba="0.5557254592689047 0.6025541199284764 0.902390111497001 1" class="collision" friction="{friction_geom}" mass="{mass}"/>
                    </body>
                  </body>"""
        scale = np_random.uniform(0.012, 0.0135)
        base_path = os.path.join(os.path.dirname(__file__), "assets/meshes")
        mesh_asset_xml = f"""
        <mesh file="{base_path}/knob/knob.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_0.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_1.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_2.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_3.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_4.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_5.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_6.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_7.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_8.obj" scale="{scale} {scale} {scale}"/>
        <mesh file="{base_path}/knob/knob_collision_9.obj" scale="{scale} {scale} {scale}"/>
        """

    elif handle_type == "pullbar":
        friction_geom = "5 5 5"
        handle_body_xml = f"""
        <body name="{handle_name}_body" pos="{pos_x} {pos_y} {pos_z}" euler="{euler_x} {euler_y} {euler_z}">
            <geom material="{handle_material}" pos="0 0 0" size="{handle_radius} {handle_radius} {handle_length}" type="box" />
            <geom name="{handle_name}" material="{handle_material}" pos="0 0 0" size="{handle_radius} {handle_radius} {handle_length}" type="box"/>
            <geom material="{handle_material}" pos="-0.04 0 {handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" mass="{mass / 10}"/>
            <geom material="{handle_material}" pos="-0.04 0 -{handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" mass="{mass / 10}"/>
            <geom material="{handle_material}" pos="-0.04 0 {handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" conaffinity="0" contype="0"/>
            <geom material="{handle_material}" pos="-0.04 0 -{handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" conaffinity="0" contype="0"/>
        </body>"""
        mesh_asset_xml = ""

    else:
        friction_geom = "5 5 5"
        handle_body_xml = f"""
        <body name="{handle_name}_body" pos="{pos_x} {pos_y} {pos_z}" euler="{euler_x} {euler_y} {euler_z}">
            <geom material="{handle_material}" pos="0 0 0" size="{handle_radius} {handle_length}" type="cylinder" friction="{friction_geom}"/>
            <geom name="{handle_name}" material="{handle_material}" pos="0 0 0" size="{handle_radius} {handle_length}" type="cylinder" mass="{mass}" friction="{friction_geom}"/>
            <geom material="{handle_material}" pos="-0.04 0 {handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" mass="{mass / 10}"/>
            <geom material="{handle_material}" pos="-0.04 0 -{handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" mass="{mass / 10}"/>
            <geom material="{handle_material}" pos="-0.04 0 {handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" conaffinity="0" contype="0"/>
            <geom material="{handle_material}" pos="-0.04 0 -{handle_length - 0.03}" euler="0 1.57 0" size="{handle_radius / 1.2} {0.04} {0.021}" type="cylinder" conaffinity="0" contype="0"/>
        </body>"""
        mesh_asset_xml = ""

    return handle_body_xml, mesh_asset_xml


def make_cabinet_xml(np_random=None):
    if np_random is None:
        np_random = np.random
    
    hinge_side = np_random.choice(["left", "right"])
    holder_type = np_random.choice(["handle", "pullbar"])
    double_doors = np_random.choice([0, 1])
    
    if double_doors:
        width = np_random.uniform(0.4, 0.8)
    else:
        width = np_random.uniform(0.2, 0.5)
    
    height = np_random.uniform(0.4, 0.8)
    depth = np_random.uniform(0.1, 0.3)
    handle_position = np_random.uniform(0.5, 0.8)
    handle_length = np_random.uniform(0.05, 0.1)
    handle_radius = np_random.uniform(0.004, 0.006)
    
    wall_thickness = 0.01
    door_gap = 0.00
    add_wall = 1
    handle_offset_y = 0.05
    handle_offset_z = 0.00
    door_mass = 0.001
    friction_geom = "3 3 3" if holder_type == "knob" else "0 0 0"
    door_gap_between = 0.005
    texrepeat_xy = (5, 5)
    paint_rgba = (1.0, 1.0, 1.0, 1.0)
    collision_rgba = (0.3, 0.3, 1.0, 0.5)

    base_path = os.path.join(os.path.dirname(__file__), "assets/textures")
    door_texture_png = np_random.choice([f"{base_path}/door/{i}.png" for i in range(2)])
    cabinet_texture_png = np_random.choice([f"{base_path}/door/{i}.png" for i in range(2)])
    handle_texture_png = np_random.choice([f"{base_path}/handle/{i}.png" for i in range(4)])
    floor_texture_png = np_random.choice([f"{base_path}/tiles/{i}.png" for i in range(3)])
    wall_texture_png = np_random.choice([f"{base_path}/wall/{i}.png" for i in range(6)])

    hx = width / 2.0
    hy = depth / 2.0
    hz = height / 2.0
    t = wall_thickness

    interior_width = width - 2 * t
    door_thickness_y = t

    if double_doors:
        door_half_x = (hx - t / 2.0) / 2.0
    else:
        door_half_x = hx - t / 2.0

    door_half_y = door_thickness_y / 2.0
    door_half_z = hz - t / 2.0

    side_x = hx - t / 2.0
    top_z = hz - t / 2.0
    bottom_z = -hz + t / 2.0
    back_y = hy - t / 2.0

    side_half_x = t / 2.0
    side_half_y = (depth - t) / 2.0
    side_half_z = hz

    tb_half_x = interior_width / 2.0
    tb_half_y = (depth - t) / 2.0
    tb_half_z = t / 2.0

    back_half_x = hx - t / 2.0
    back_half_y = t / 2.0
    back_half_z = hz - t / 2.0

    repx, repy = texrepeat_xy
    wood_texrepeat = f"{repx} {repy}"

    paint_r, paint_g, paint_b, paint_a = paint_rgba
    col_r, col_g, col_b, col_a = collision_rgba

    wall_thickness_back = 0.02
    wall_y_pos = hy + wall_thickness_back / 2.0

    wall_section = ""
    if add_wall:
        wall_section = f"""
        <!-- Large wall behind cabinet -->
        <body name="front_wall">    
                <geom name="wall" pos="0 {wall_y_pos} 0" size="0 0 0.05" euler="1.57 0 0" type="plane" material="M_wall"/>
          </body>
        """

    def build_holder(
        hinge_side,
        holder_type,
        handle_x,
        handle_y,
        handle_z,
        friction_geom,
        door_mass,
        handle_radius,
        handle_length,
        handle_name,
    ):
        # Use the new reusable handle function
        handle_body_xml, mesh_asset_xml = make_handle_xml(
            handle_type=holder_type,
            handle_name=handle_name,
            position=(handle_x, handle_y, handle_z),
            euler_orientation=(0, 0, -1.57),
            handle_radius=handle_radius,
            handle_length=handle_length,
            handle_material="M_hinge_metal",
            mass=0.1,
            friction_geom=friction_geom,
            hinge_side=hinge_side,
            np_random=np_random,
        )

        return handle_body_xml, mesh_asset_xml

    def make_single_door(
        name_suffix, x_offset, hinge_side, joint_name, handle_name, door_width_half
    ):
        hinge_sign = +1.0 if hinge_side.lower() == "right" else -1.0

        # For double doors, adjust hinge position so doors are flush with cabinet sides and add gap
        if double_doors:
            door_side_offset = hx - t / 2.0
            gap = door_gap_between / 2.0
            if hinge_side.lower() == "left":
                door_root_x = -door_side_offset - gap
            else:
                door_root_x = door_side_offset + gap
        else:
            door_root_x = hinge_sign * side_x

        door_root_z = 0.0
        door_center_x = -hinge_sign * door_width_half
        door_center_y = -(t / 2.0)

        if double_doors:
            handle_x = door_center_x - hinge_sign * (door_width_half - 0.06)
        else:
            handle_x = door_center_x - hinge_sign * (door_width_half - 0.05)
        handle_y = -door_half_y - handle_offset_y
        door_bottom = -door_half_z
        door_top = door_half_z
        handle_z = (
            door_bottom + handle_position * (door_top - door_bottom) + handle_offset_z
        )

        if hinge_side.lower() == "right":
            jmin, jmax = 0.0, 1.57
        else:
            jmin, jmax = -1.57, 0.0

        holder_xml, mesh_asset = build_holder(
            hinge_side,
            holder_type,
            handle_x,
            handle_y,
            handle_z,
            friction_geom,
            door_mass,
            handle_radius,
            handle_length,
            handle_name,
        )

        return (
            f"""
            <body name="hingedoor{name_suffix}" pos="{door_root_x} {-hy - door_gap} {door_root_z}">
                <joint axis="0 0 1" name="{joint_name}" range="{jmin} {jmax}"  damping="2" frictionloss="2" armature=".01"/>
                {holder_xml}
                <body name="door_body{name_suffix}">
                    <geom material="M_door_wood" pos="{door_center_x} {door_center_y} 0" size="{door_width_half} {door_half_z} {door_half_y}" euler="1.57 0 0" type="box"/>
                </body>
                <site type="sphere" name="boxdoor_site{name_suffix}" pos="{handle_x} {handle_y - 0.018} {handle_z}" size=".01" group="3" rgba="1 0 0 1"/>
            </body>
        """,
            mesh_asset,
        )

    def make_single_cabinet(name_suffix, x_offset, hinge_side, joint_name, handle_name):
        hinge_sign = +1.0 if hinge_side.lower() == "right" else -1.0
        door_root_x = hinge_sign * side_x
        door_root_z = 0.0
        door_center_x = -hinge_sign * (door_half_x)
        door_center_y = -(t / 2.0)

        handle_x = door_center_x - hinge_sign * (door_half_x - 0.05)
        handle_y = -door_half_y - handle_offset_y
        door_bottom = -door_half_z
        door_top = door_half_z
        handle_z = (
            door_bottom + handle_position * (door_top - door_bottom) + handle_offset_z
        )

        if hinge_side.lower() == "right":
            jmin, jmax = 0.0, 1.57
        else:
            jmin, jmax = -1.57, 0.0

        holder_xml, mesh_asset = build_holder(
            hinge_side,
            holder_type,
            handle_x,
            handle_y,
            handle_z,
            friction_geom,
            door_mass,
            handle_radius,
            handle_length,
            handle_name,
        )

        return (
            f"""
        <body name="cabinet_object{name_suffix}" childclass="boxcabinet" pos="{x_offset} 0 {hz}">
            <body name="back_panel{name_suffix}">
                <geom material="M_cabinet_paint" pos="-{side_x} {-t / 4.0} 0" size="{side_half_x} {side_half_y} {side_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="{side_x} {-t / 4.0} 0" size="{side_half_x} {side_half_y} {side_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="0 {-t / 4.0} {top_z}" size="{tb_half_x} {tb_half_y} {tb_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="0 {-t / 4.0} {bottom_z}" size="{tb_half_x} {tb_half_y} {tb_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="0 {back_y} 0" size="{back_half_x} {back_half_y} {back_half_z}" type="box" conaffinity="0" contype="0"/>
            </body>

            <body name="hingedoor{name_suffix}" pos="{door_root_x} {-hy - door_gap} {door_root_z}">
                <joint axis="0 0 1" name="{joint_name}" range="{jmin} {jmax}" damping="2" frictionloss="2" armature=".01"/>
                {holder_xml}
                <body name="door_body{name_suffix}">
                    <geom material="M_door_wood" pos="{door_center_x} {door_center_y} 0" size="{door_half_x} {door_half_z} {door_half_y}" euler="1.57 0 0" type="box"/>
                </body>
                <site type="sphere" name="boxdoor_site{name_suffix}" pos="{handle_x} {handle_y - 0.018} {handle_z}" size=".01" group="3" rgba="1 0 0 1"/>
            </body>
        </body>
        """,
            mesh_asset,
        )

    cabinets_xml, extra_meshes = "", ""

    if double_doors:
        # Randomly pick which door gets 'task_joint' and 'handle_1'
        if np_random.uniform() > 0.5:
            left_joint = "task_joint"
            right_joint = "task_joint_2"
        else:
            left_joint = "task_joint_2"
            right_joint = "task_joint"

        left_handle = "handle" if left_joint == "task_joint" else "handle_2"
        right_handle = "handle" if right_joint == "task_joint" else "handle_2"

        left_door_xml, mesh1 = make_single_door(
            "_L", -door_half_x, "left", left_joint, left_handle, door_half_x
        )
        right_door_xml, mesh2 = make_single_door(
            "_R", door_half_x, "right", right_joint, right_handle, door_half_x
        )

        cabinets_xml = f"""
        <body name="cabinet_object" childclass="boxcabinet" pos="0 0 {hz}">
            <body name="back_panel">
                <geom material="M_cabinet_paint" pos="-{side_x} {-t / 4.0} 0" size="{side_half_x} {side_half_y} {side_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="{side_x} {-t / 4.0} 0" size="{side_half_x} {side_half_y} {side_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="0 {-t / 4.0} {top_z}" size="{tb_half_x} {tb_half_y} {tb_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="0 {-t / 4.0} {bottom_z}" size="{tb_half_x} {tb_half_y} {tb_half_z}" type="box" conaffinity="0" contype="0"/>
                <geom material="M_cabinet_paint" pos="0 {back_y} 0" size="{back_half_x} {back_half_y} {back_half_z}" type="box" conaffinity="0" contype="0"/>
            </body>
            {left_door_xml}
            {right_door_xml}
        </body>
        """
        extra_meshes = mesh1 + mesh2

        exclude_doors_xml = """
        <contact> <exclude body1="hingedoor_L" body2="hingedoor_R"/> <exclude body1="door_body_L" body2="door_body_R"/>
        
        <exclude body1="left_collisions" body2="door_body_L"/>
        <exclude body1="left_collisions" body2="door_body_R"/>
        <exclude body1="right_collisions" body2="door_body_L"/>
        <exclude body1="right_collisions" body2="door_body_R"/>

        </contact>

        """
    else:
        cabinets_xml, extra_meshes = make_single_cabinet(
            "", 0, hinge_side, "task_joint", "handle"
        )
        exclude_doors_xml = """
        <contact>
        <exclude body1="left_collisions" body2="door_body"/>
        <exclude body1="right_collisions" body2="door_body"/>



        </contact>

        """

     
    xml = f"""
            <mujoco model="hinge cabinet">
            <compiler angle="radian"/>
                <asset>
                    {extra_meshes}
                    <texture name="T_door_wood" type="2d" file="{door_texture_png}"/>
                    <texture name="T_cabinet_wood" type="2d" file="{cabinet_texture_png}"/>
                    <texture name="T_hinge_metal" type="2d" file="{handle_texture_png}"/>
                    <texture name="T_floor" type="2d" file="{floor_texture_png}"/>
                    <texture name="T_wall" type="2d" file="{wall_texture_png}"/>
                    <material name="M_door_wood" texture="T_door_wood" texrepeat="5 5" reflectance="0.0" shininess=".1" texuniform="true"/>
                    <material name="M_cabinet_wood" texture="T_cabinet_wood" texrepeat="{wood_texrepeat}" reflectance="0.0" shininess=".1" texuniform="true"/>
                    <material name="M_hinge_metal" texture="T_hinge_metal" texrepeat="4 4" reflectance="0.6" shininess=".6" texuniform="true"/>
                    <material name="M_floor" texture="T_floor" texrepeat="3 3" reflectance="0.0" shininess=".1" texuniform="true"/>
                    <material name="M_wall" texture="T_wall" texrepeat="4 4" reflectance="0.0" shininess=".1" texuniform="true"/>
                    <material name="M_cabinet_paint" texture="T_cabinet_wood" rgba="{paint_r} {paint_g} {paint_b} {paint_a}" reflectance="0.0" shininess=".01" texuniform="true"/>
                    <material name="hinge_collision_blue" rgba="{col_r} {col_g} {col_b} {col_a}" shininess="0" specular="0"/>
                </asset>
                <default>
                    <default class="boxcabinet">
                        <joint damping="2" frictionloss="2" limited="true"/>
                        <geom group="1" material="M_cabinet_wood" type="mesh"/>
                        <default class="box_collision">
                            <geom conaffinity="1" condim="6" contype="1" group="4" material="hinge_collision_blue"/>
                        </default>
                    </default>
                </default>
                <worldbody>
                    {wall_section}
                    {cabinets_xml}
                
                    <body name="floor">
                        <geom name="floor" pos="0 0 0" size="0 0 0.05" euler="0 0 0" type="plane" material="M_floor"/>
                    </body>
                </worldbody>
                {exclude_doors_xml}
            </mujoco>
            """
    return xml

def make_drawer_xml(np_random=None):
    if np_random is None:
        np_random = np.random
    
    handle_type = np_random.choice(["handle", "pullbar"])
    height_per_drawer = np_random.uniform(0.05, 0.15)
    handle_radius = np_random.uniform(0.005, 0.01)
    width = np_random.uniform(0.2, 0.4)
    depth = np_random.uniform(0.15, 0.3)
    handle_length = np_random.uniform(0.07, max(0.05, width / 1.2))
    num_drawers = np_random.integers(1, 5)
    
    position = (0, 0, 0)
    rotation = (0, 0, -1.57)
    handle_protrusion = 0.05
    thickness = 0.005
    cover_color = (0.3, 0.2, 0.05, 1.0)
    drawer_color = (0.4, 0.25, 0.1, 1.0)
    handle_color = (0.2, 0.2, 0.2, 1.0)
    texrepeat_xy = (3, 3)
    add_wall = 1
    joint_damping = 2

    base_path = os.path.join(os.path.dirname(__file__), "assets/textures")
    cover_texture_png = np_random.choice([f"{base_path}/door/{i}.png" for i in range(3)])
    drawer_texture_png = np_random.choice([f"{base_path}/door/{i}.png" for i in range(3)])
    door_texture_png = np_random.choice([f"{base_path}/door/{i}.png" for i in range(3)])
    handle_texture_png = np_random.choice([f"{base_path}/handle/{i}.png" for i in range(4)])
    floor_texture_png = np_random.choice([f"{base_path}/tiles/{i}.png" for i in range(3)])
    wall_texture_png = np_random.choice([f"{base_path}/wall/{i}.png" for i in range(6)])

    half_total_height = num_drawers * height_per_drawer + thickness * (num_drawers + 1)
    total_height = half_total_height * 2

    adjusted_position = (position[0], position[1], thickness)

    repx, repy = texrepeat_xy
    texrepeat_str = f"{repx} {repy}"

    cover_r, cover_g, cover_b, cover_a = cover_color
    drawer_r, drawer_g, drawer_b, drawer_a = drawer_color
    handle_r, handle_g, handle_b, handle_a = handle_color

    base_template = f"""<mujoco model="drawer">
    <asset>
        <texture name="T_cover" type="2d" height="1" width="1" file="{cover_texture_png}"/>
        <texture name="T_drawer" type="2d" height="1" width="1" file="{drawer_texture_png}"/>
        <texture name="T_door" type="2d" height="1" width="1" file="{door_texture_png}"/>
        <texture name="T_handle" type="2d" height="1" width="1" file="{handle_texture_png}"/>
        <texture name="T_floor" type="2d" height="2" width="2" file="{floor_texture_png}"/>
        <texture name="T_wall" type="2d" height="2" width="1" file="{wall_texture_png}"/>
    
    <material name="M_cover" texture="T_cover" texrepeat="{texrepeat_str}" reflectance="0.0" shininess=".1" texuniform="false"/>
    <material name="M_drawer" texture="T_drawer" texrepeat="{texrepeat_str}" reflectance="0.0" shininess=".1" texuniform="false"/>
    <material name="M_door" texture="T_door" texrepeat="{texrepeat_str}" reflectance="0.0" shininess=".1" texuniform="false"/>
    <material name="M_handle" texture="T_handle" texrepeat="{texrepeat_str}" reflectance="0.0" shininess=".1" texuniform="false"/>
    <material name="M_floor" texture="T_floor" texrepeat="5 5" reflectance="0.0" shininess=".1" texuniform="false"/>
    <material name="M_wall" texture="T_wall" texrepeat="2 2" reflectance="0.0" shininess=".1" texuniform="false"/>
  </asset>

    <worldbody>
      
        <body name="drawer_object" pos="{adjusted_position[0]} {adjusted_position[1]} {adjusted_position[2]}" euler="{rotation[0]} {rotation[1]} {rotation[2]}">
            <body name="outer_frame">
      <geom size="{depth + thickness * 3} {thickness} {half_total_height}" pos="0 -{width + thickness * 2} {half_total_height}" type="box" material="M_cover" />
      <geom size="{depth + thickness * 3} {thickness} {half_total_height}" pos="0 {width + thickness * 2} {half_total_height}" type="box" material="M_cover" />
      <geom size="{depth + thickness * 3} {width + thickness * 3} {thickness}" pos="0 0 {thickness}" type="box" material="M_cover" />
      <geom size="{depth + thickness * 3} {width + thickness * 3} {thickness}" pos="0 0 {total_height}" type="box" material="M_cover" />
      <geom size="{thickness} {width + thickness * 3} {half_total_height}" pos="-{depth + thickness * 2} 0 {half_total_height}" type="box" material="M_cover" />

        </body>
      """

    drawer_bodies = ""

    task_drawer_index = np_random.integers(0, num_drawers)

    for i in range(num_drawers):
        drawer_name = f"drawer_{i}"

        if i == task_drawer_index:
            joint_name = "task_joint"
            handle_geom_name = "handle"
        else:
            joint_name = f"drawer_{i}_slide"
            if i == 1 and task_drawer_index != 1:
                handle_geom_name = "handle_1_normal"
            else:
                handle_geom_name = f"handle_{i}"

        z_pos = height_per_drawer * (2 * i + 1) + thickness * (2 * i + 3)
        collision_thickness = thickness * 0.8
        collision_size_adjust = thickness * 0.8

        drawer_body = f"""
      <body name="{drawer_name}" pos="{thickness * 3} 0 {z_pos}">
        <joint name="{joint_name}" type="slide" axis="1 0 0" limited="true" range="0 {
            depth * 2
        }"  damping="10" frictionloss="10" armature=".01"   />
        
        <geom size="{depth} {thickness} {height_per_drawer}" type="box" pos="0 {
            width - thickness
        } 0" material="M_drawer" conaffinity="0" contype="0"/>
        <geom size="{depth} {thickness} {height_per_drawer}" type="box" pos="0 -{
            width - thickness
        } 0" material="M_drawer" conaffinity="0" contype="0"/>
        <geom size="{height_per_drawer} {width} {thickness}" type="box" pos="{
            depth - thickness
        } 0 0" material="M_drawer" conaffinity="0" contype="0" euler="1.57 1.57 1.57"/>
        <geom size="{height_per_drawer} {width} {thickness}" type="box" pos="{
            depth - thickness
        } 0 0" material="M_drawer" conaffinity="0" contype="0" euler="1.57 1.57 1.57"/>
          <geom size="{height_per_drawer} {width} {thickness}" type="box" pos="{
            depth - thickness
        } 0 0" material="M_drawer" euler="1.57 1.57 1.57"/>
        
        <geom size="{depth * 0.9} {width * 0.9} {
            collision_thickness
        }" type="box" pos="0 0 -{height_per_drawer - thickness}" mass="0.0001"/>
        <geom size="{depth * 0.9} {collision_size_adjust} {
            height_per_drawer * 0.8
        }" type="box" pos="0 {width - thickness} 0" mass="0.0001"/>
        <geom size="{depth * 0.9} {collision_size_adjust} {
            height_per_drawer * 0.8
        }" type="box" pos="0 -{width - thickness} 0" mass="0.0001"/>
        <geom size="{collision_size_adjust} {width * 0.9} {
            height_per_drawer * 0.8
        }" type="box" pos="{depth - thickness} 0 0" mass="0.0001"/>
        <geom size="{collision_size_adjust} {width * 0.9} {
            height_per_drawer * 0.8
        }" type="box" pos="-{depth - thickness} 0 0" mass="0.0001"/>
        
        {
            make_handle_xml(
                handle_type=handle_type,
                handle_name=handle_geom_name,
                position=(depth + handle_protrusion, 0, 0),
                euler_orientation=(1.57, 0, 0),
                handle_radius=handle_radius,
                handle_length=handle_length,
                handle_material="M_handle",
                mass=0.1,
                friction_geom="0 0 0",
            )[0]
        }
      </body>"""

        drawer_bodies += drawer_body

    wall_y_pos = -(depth + 0.55)

    floor_wall_section = """
        <body name="floor">
            <geom name="floor" pos="0 0 0" size="0 0 0.05" euler="0 0 0" type="plane" material="M_floor" conaffinity="0" contype="0"/>
        </body>"""

    if add_wall:
        floor_wall_section += f"""
        <!-- Large wall behind drawer -->
        <body name="front_wall">    
                <geom name="wall" pos="0 {-wall_y_pos} 0" size="0 0 0.05" euler="1.57 0 0" type="plane" material="M_wall"/>
        </body>
       """

    # Dynamically generate contact exclusions based on num_drawers
    contact_excludes = ""
    for i in range(num_drawers):
        contact_excludes += (
            f'        <exclude body1="drawer_{i}" body2="outer_frame"/>\n'
        )

    closing_template = (
        """
    </body>"""
        + floor_wall_section
        + f"""
  </worldbody>

  <contact>
{contact_excludes}  </contact>
</mujoco>"""
    )

    drawer_xml_string = base_template + drawer_bodies + closing_template
    return drawer_xml_string

def resize_with_pad(images, height, width):
    if images.shape[-3:-1] == (height, width):
        return images

    original_shape = images.shape
    images = images.reshape(-1, *original_shape[-3:])
    resized = np.stack(
        [
            _resize_with_pad_pil(Image.fromarray(im), height, width, method=Image.BILINEAR)
            for im in images
        ]
    )
    return resized.reshape(*original_shape[:-3], *resized.shape[-3:])


def _resize_with_pad_pil(image, height, width, method):
    cur_width, cur_height = image.size
    if cur_width == width and cur_height == height:
        return image

    ratio = max(cur_width / width, cur_height / height)
    resized_height = int(cur_height / ratio)
    resized_width = int(cur_width / ratio)
    resized_image = image.resize((resized_width, resized_height), resample=method)

    zero_image = Image.new(resized_image.mode, (width, height), 0)
    pad_height = max(0, int((height - resized_height) / 2))
    pad_width = max(0, int((width - resized_width) / 2))
    zero_image.paste(resized_image, (pad_width, pad_height))
    assert zero_image.size == (width, height)
    return zero_image

def set_table_size(scene_xml_content, table_width, table_length):
    return scene_xml_content.replace("{TABLE_WIDTH}", str(table_width)).replace(
        "{TABLE_LENGTH}", str(table_length)
    )

def pixel_to_world(u, v, depth, model, data, cam_name, img_width, img_height):
    fovy = math.radians(model.camera(cam_name).fovy[0])
    f = img_height / (2 * np.tan(fovy / 2))
    cam_mat = np.array([[f, 0, img_width / 2], [0, f, img_height / 2], [0, 0, 1]])

    cx = cam_mat[0, 2]
    cy = cam_mat[1, 2]
    fx = cam_mat[0, 0]
    fy = cam_mat[1, 1]

    z_cam = -depth
    x_cam = -(u - cx) * z_cam / fx
    y_cam = -(cy - v) * z_cam / fy

    p_cam = np.array([x_cam, y_cam, z_cam])

    xmat = data.camera(cam_name).xmat
    R_wc = xmat.reshape(3, 3)
    pos = data.camera(cam_name).xpos
    p_world = R_wc @ p_cam + pos
    return p_world


import gymnasium as gym
from .tasks.pick import PickTask
from .tasks.open import OpenTask
from .tasks.close import CloseTask

gym.register(id="Egogym-Pick-v0", entry_point=PickTask)
gym.register(id="Egogym-Open-v0", entry_point=OpenTask)
gym.register(id="Egogym-Close-v0", entry_point=CloseTask)


import argparse
import asyncio
import json
import logging
import re
import numpy as np
import torch
import websockets
from PIL import Image
from transformers import AutoModelForCausalLM, AutoProcessor, GenerationConfig

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def extract_molmo_points(molmo_output: str):
    points = []
    for match in re.finditer(
        r'x\d*="\s*([0-9]+(?:\.[0-9]+)?)"\s+y\d*="\s*([0-9]+(?:\.[0-9]+)?)"',
        molmo_output,
    ):
        try:
            p = np.array([float(match.group(1)), float(match.group(2))], dtype=np.float32)
        except ValueError:
            continue

        if np.max(p) > 100:
            continue

        points.append(p / 100.0)

    return points


class Molmo:
    def __init__(self, model_name="allenai/Molmo-7B-D-0924"):
        self.processor = AutoProcessor.from_pretrained(
            model_name,
            trust_remote_code=True,
            torch_dtype="auto",
            device_map="auto",
        )
        self.model = AutoModelForCausalLM.from_pretrained(
            model_name,
            trust_remote_code=True,
            torch_dtype="auto",
            device_map="auto",
        )

    def infer(self, rgb: np.ndarray, prompt: str) -> str:
        image = Image.fromarray(rgb)
        inputs = self.processor.process(images=[image], text=prompt)
        inputs = {k: v.to(self.model.device).unsqueeze(0) for k, v in inputs.items()}

        with torch.inference_mode():
            output = self.model.generate_from_batch(
                inputs,
                GenerationConfig(max_new_tokens=200, stop_strings="<|endoftext|>"),
                tokenizer=self.processor.tokenizer,
            )

        gen_tokens = output[0, inputs["input_ids"].size(1):]
        return self.processor.tokenizer.decode(gen_tokens, skip_special_tokens=True)

    def infer_point(self, rgb: np.ndarray, prompt: str) -> np.ndarray:
        text = self.infer(rgb, prompt)
        log.info(f"Molmo output: {text}")
        points = extract_molmo_points(text)
        return points[0] if points else np.array([0.5, 0.5], dtype=np.float32)


class MolmoWebSocketServer:
    def __init__(self, host="0.0.0.0", port=8765):
        self.host = host
        self.port = port
        self.molmo = Molmo()

    async def handle_client(self, websocket):
        client_id = id(websocket)
        log.info(f"Client connected: {client_id}")

        try:
            async for msg in websocket:
                try:
                    req = json.loads(msg)
                    action = req.get("action")

                    if action != "infer_point":
                        raise ValueError("Only 'infer_point' action is supported")

                    rgb = np.array(req["rgb"], dtype=np.uint8)

                    if "object_name" in req:
                        prompt = f"Point to the center of the {req['object_name']}."
                        label = req["object_name"].replace(" ", "_")
                    elif "prompt" in req:
                        prompt = req["prompt"]
                        label = "custom_prompt"
                    else:
                        raise ValueError("Provide either 'object_name' or 'prompt'")

                    point = self.molmo.infer_point(rgb, prompt)

                    resp = {
                        "status": "ok",
                        "action": "infer_point",
                        "point": point.tolist(),
                    }

                except Exception as e:
                    log.exception("Request error")
                    resp = {"status": "error", "message": str(e)}

                await websocket.send(json.dumps(resp))

        except websockets.exceptions.ConnectionClosed:
            log.info(f"Client disconnected: {client_id}")

    async def start(self):
        log.info(f"Starting Molmo server on ws://{self.host}:{self.port}")
        async with websockets.serve(
            self.handle_client,
            self.host,
            self.port,
            max_size=50 * 1024 * 1024,
        ):
            await asyncio.Future()


def main():
    parser = argparse.ArgumentParser("Molmo Pointing Server")
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", type=int, default=8765)
    args = parser.parse_args()

    server = MolmoWebSocketServer(
        host=args.host,
        port=args.port,
    )
    asyncio.run(server.start())


if __name__ == "__main__":
    main()



import json
import numpy as np
import websockets

class MolmoClient:
    def __init__(self, host="localhost", port=8765):
        self.uri = f"ws://{host}:{port}"
        self.websocket = None

    async def connect(self):
        self.websocket = await websockets.connect(
            self.uri,
            max_size=50 * 1024 * 1024,
        )

    async def disconnect(self):
        if self.websocket:
            await self.websocket.close()

    async def infer_point(self, rgb: np.ndarray, object_name: str = None, prompt: str = None):
        if not self.websocket:
            raise RuntimeError("Not connected to server. Call connect() first.")

        req = {
            "action": "infer_point",
            "rgb": rgb.tolist(),
        }

        if object_name:
            req["object_name"] = object_name
        elif prompt:
            req["prompt"] = prompt
        else:
            raise ValueError("Provide either 'object_name' or 'prompt'")

        await self.websocket.send(json.dumps(req))
        resp = json.loads(await self.websocket.recv())

        if resp["status"] == "error":
            raise RuntimeError(f"Server error: {resp['message']}")

        return np.array(resp["point"], dtype=np.float32)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()



import torch
import numpy as np
import cv2
from PIL import Image
from scipy.spatial.transform import Rotation as R
import torchvision.transforms as T
from hydra import initialize, compose
import hydra

P = np.array(
    [[-1, 0, 0, 0], [0, 0, -1, 0], [0, -1, 0, 0], [0, 0, 0, 1]], dtype=np.float32
)

def init_model_loss_fn(cfg):
    device = cfg.device if torch.cuda.is_available() else "cpu"
    model = hydra.utils.instantiate(cfg.model).to(device)
    model_weight_pth = cfg.get("model_weight_pth")

    if model_weight_pth is None:
        raise ValueError("Model weight path is not specified in the config.")

    checkpoint = torch.load(
        model_weight_pth, map_location=device, weights_only=False
    )

    try:
        model.load_state_dict(checkpoint["model"])
    except RuntimeError:
        checkpoint["model"] = {
            k.replace("_orig_mod.", ""): v for k, v in checkpoint["model"].items()
        }
        model.load_state_dict(checkpoint["model"])
    loss_fn = hydra.utils.instantiate(cfg.loss_fn, model=model)
    loss_fn.load_state_dict(checkpoint["loss_fn"])
    loss_fn = loss_fn.to(device)

    return model, loss_fn

class VectorizedBuffer:
    def __init__(self, batch_size, buffer_size, act_dim, device):
        self.batch_size = batch_size
        self.buffer_size = buffer_size
        self.act_dim = act_dim
        self.device = device
        self.image_buffer = None
        self.goal_buffer = None
        self.action_buffer = None
        self.image_buffers_sizes = torch.zeros(batch_size, device=device)
        self.goal_buffer_size = torch.zeros(batch_size, device=device)
        self.action_buffers_sizes = torch.zeros(batch_size, device=device)

    def add_image(self, new_images):
        if self.image_buffer is None:
            self.image_buffer = (
                new_images.unsqueeze(1)
                .repeat(1, self.buffer_size, 1, 1, 1)
                .to(self.device)
            )
        else:
            for b in range(self.batch_size):
                if self.image_buffers_sizes[b] == 0:
                    self.image_buffer[b] = (
                        new_images[b].unsqueeze(0).repeat(self.buffer_size, 1, 1, 1)
                    )
                else:
                    self.image_buffer[b] = torch.roll(
                        self.image_buffer[b], shifts=-1, dims=0
                    )
                    self.image_buffer[b, -1] = new_images[b]
        self.image_buffers_sizes += 1
        self.image_buffers_sizes = torch.clamp(
            self.image_buffers_sizes, max=self.buffer_size
        )

    def add_goal(self, new_goals):
        if self.goal_buffer is None:
            self.goal_buffer = (
                new_goals.unsqueeze(1).repeat(1, self.buffer_size, 1).to(self.device)
            )
        else:
            for b in range(self.batch_size):
                if self.goal_buffer_size[b] == 0:
                    self.goal_buffer[b] = (
                        new_goals[b].unsqueeze(0).repeat(self.buffer_size, 1)
                    )
                else:
                    self.goal_buffer[b] = torch.roll(
                        self.goal_buffer[b], shifts=-1, dims=0
                    )
                    self.goal_buffer[b, -1] = new_goals[b]
        self.goal_buffer_size += 1
        self.goal_buffer_size = torch.clamp(self.goal_buffer_size, max=self.buffer_size)

    def reset(self, batch_indices):
        self.image_buffers_sizes[batch_indices] = 0
        self.goal_buffer_size[batch_indices] = 0
        self.action_buffers_sizes[batch_indices] = 0
        if self.image_buffer is not None:
            self.image_buffer[batch_indices] = torch.zeros_like(
                self.image_buffer[batch_indices]
            )
        if self.goal_buffer is not None:
            self.goal_buffer[batch_indices] = torch.zeros_like(
                self.goal_buffer[batch_indices]
            )
        if self.action_buffer is not None:
            self.action_buffer[batch_indices] = torch.zeros_like(
                self.action_buffer[batch_indices]
            )

    def add_action(self, new_actions):
        B = new_actions.shape[0]
        if self.action_buffer is None:
            self.action_buffer = torch.zeros(
                B, self.buffer_size - 1, self.act_dim, device=self.device
            )
            self.action_buffers_sizes = torch.zeros(B, device=self.device)
        for b in range(B):
            if self.action_buffers_sizes[b] != 0:
                self.action_buffer[b] = torch.roll(
                    self.action_buffer[b], shifts=-1, dims=0
                )
            self.action_buffer[b, -1] = new_actions[b]
        self.action_buffers_sizes += 1

    def get_input_sequence(self):
        B = self.image_buffer.shape[0]
        if self.action_buffer is None:
            action_buffer = torch.zeros(
                B, self.buffer_size - 1, self.act_dim, device=self.device
            )
        else:
            action_buffer = self.action_buffer
        base_act = torch.zeros(B, 1, self.act_dim, device=self.device)
        act_seq = torch.cat([action_buffer, base_act], dim=1)
        if self.goal_buffer is None:
            goal_seq = None
        else:
            goal_seq = torch.stack([goal for goal in self.goal_buffer]).to(
                dtype=torch.float32
            )
        return self.image_buffer, goal_seq, act_seq


def unwrap_model(model):
    if isinstance(
        model, (torch.nn.DataParallel, torch.nn.parallel.DistributedDataParallel)
    ):
        return model.module
    return model


class Policy:
    def __init__(self, model_path=None, device="cpu", model=None, loss_fn=None):
        if model_path is None and (model is None or loss_fn is None):
            raise ValueError(
                "Either model_path or both model and loss_fn must be provided."
            )

        if model is None or loss_fn is None:
            with initialize(config_path="configs", version_base=None):
                cfg = compose(config_name="run_vqbet")
            checkpoint = torch.load(model_path, map_location="cpu", weights_only=False)
            ckpt_cfg = checkpoint["cfg"]
            del checkpoint

            cfg.goal_dim = ckpt_cfg["loss_fn"]["goal_dim"]

            cfg.gpt_input_dim = ckpt_cfg["loss_fn"]["gpt_model"]["config"]["input_dim"]
            cfg.loss_fn.gpt_model.config.n_layer = ckpt_cfg["loss_fn"]["gpt_model"][
                "config"
            ]["n_layer"]
            cfg.loss_fn.gpt_model.config.n_head = ckpt_cfg["loss_fn"]["gpt_model"][
                "config"
            ]["n_head"]
            cfg.loss_fn.gpt_model.config.n_embd = ckpt_cfg["loss_fn"]["gpt_model"][
                "config"
            ]["n_embd"]

            cfg.vqvae_n_embed = ckpt_cfg["vqvae_n_embed"]
            cfg.model_weight_pth = model_path

            cfg.device = device
            model, loss_fn = init_model_loss_fn(cfg)

        self.to_tensor = T.ToTensor()
        self.model = unwrap_model(model)
        self.loss_fn = unwrap_model(loss_fn)
        self.buffer_size = self.loss_fn._vqbet.obs_window_size
        self.device = device

        goal_dim = self.loss_fn.goal_dim
        self.condition = f"{goal_dim}d"
        if goal_dim == 0:
            self.condition = None

        valid_conditions = ("4d", "3d", "2d")
        if self.condition is not None and self.condition not in valid_conditions:
            raise ValueError(
                f"'condition' must be one of {valid_conditions}, got '{self.condition}'"
            )

        self.model.eval()
        self.loss_fn.eval()

        self.vectorized_buffer = None
        self.act_dim = 7

        self.rot_yx_90 = (
            R.from_euler("y", 90, degrees=True).as_matrix()
            @ R.from_euler("x", 90, degrees=True).as_matrix()
        )
        self.Tyx = np.eye(4, dtype=np.float32)
        self.Tyx[:3, :3] = self.rot_yx_90

        rot_z_90 = R.from_euler("z", 90, degrees=True).as_matrix()
        Tz = np.eye(4, dtype=np.float32)
        Tz[:3, :3] = rot_z_90

        M = self.Tyx @ Tz @ P.T
        M_inv = M.T

        self.M_t = torch.from_numpy(M).to(device)
        self.M_inv_t = torch.from_numpy(M_inv).to(device)

    def reset(self, indicies=None):
        if indicies is None:
            self.vectorized_buffer = None
        else:
            self.vectorized_buffer.reset(indicies)

    def process_image(self, img):
        if isinstance(img, np.ndarray):
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            img = Image.fromarray(img)
        return self.to_tensor(img)
    
    def process_image_batch(self, imgs):
        if isinstance(imgs, np.ndarray):
            if imgs.dtype == np.uint8:
                imgs_tensor = torch.from_numpy(imgs).permute(0, 3, 1, 2).float() / 255.0
            else:
                imgs_tensor = torch.from_numpy(imgs).permute(0, 3, 1, 2).float()
            return imgs_tensor.to(self.device)
        else:
            return torch.stack([self.process_image(imgs[i]) for i in range(len(imgs))]).to(self.device)

    def infer(self, observations):
        obs = observations["rgb_ego"]
        if self.condition == "2d":
            goal = observations["object_2d_position"]
        elif self.condition == "3d":
            goal = observations["object_3d_position"]
        else:
            goal = None

        if len(obs.shape) == 3:
            obs = np.expand_dims(obs, axis=0)
            if goal is not None:
                goal = np.expand_dims(goal, axis=0)

        B = obs.shape[0]
        processed_images = self.process_image_batch(obs)

        if self.vectorized_buffer is None:
            image_shape = processed_images.shape[1:]
            self.vectorized_buffer = VectorizedBuffer(
                batch_size=B,
                buffer_size=self.buffer_size,
                act_dim=self.act_dim,
                device=self.device,
            )

        self.vectorized_buffer.add_image(processed_images)

        if self.condition is not None:
            processed_goals = torch.from_numpy(goal).to(self.device, dtype=torch.float32).view(B, -1)
            self.vectorized_buffer.add_goal(processed_goals)

        img_seq, goal_seq, act_seq = self.vectorized_buffer.get_input_sequence()

        with torch.no_grad():
            model_input = (img_seq, goal_seq, act_seq)
            model_output = self.model(model_input)
            action_tensors, logs = self.loss_fn.step(
                model_input, model_output, return_all=True
            )

        action_tensors = action_tensors.squeeze(1).to(self.device)
        self.vectorized_buffer.add_action(action_tensors)
        action_tensors = action_tensors.cpu().numpy()

        return action_tensors



import os
from gymnasium.spaces import Box, Text
import numpy as np
from scipy.spatial.transform import Rotation as R

from egogym.egogym import Egogym
from egogym.utils import include_in_scene, position_sampler, make_objects_manager, make_textures_manager
import egogym.assets.constants as constants

class PickTask(Egogym):

    def __init__(self, robot="rum", action_space="delta", render_mode=None, render_size=(960,720), num_objs=1, seed=None, objects_set=None):
        super().__init__(robot=robot, action_space=action_space, render_mode=render_mode, render_size=render_size, seed=seed)
        self.spread = 0.22
        self.num_objs = num_objs
        if objects_set is not None:
            self.objects_manager = make_objects_manager(objects_set, self.np_random, shuffle=True)
        else:
            self.objects_manager = make_objects_manager(constants.lite_pick_objects_set, self.np_random)
        self.textures_manager = make_textures_manager([f"wood/{i}.png" for i in range(10)], self.np_random)
        self.observation_space["object_pose"] = Box(low=-np.inf, high=np.inf, shape=(16,), dtype=np.float32)
        self.observation_space["object_name"] = Text(max_length=256)


    def make_task_scene(self):
        assets_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "assets"))
        
        with open(os.path.join(assets_dir, "scenes", "pick.xml"), "r") as f:
            scene_xml = f.read()

        self.object = self.objects_manager.sample()
        self.secondary_objects = []
        used_objects = {self.object.name}
        scene_xml = self.object.add_to_scene_xml(scene_xml)
        while len(used_objects) < self.num_objs:
            object = self.objects_manager.sample(random=True)
            if object.name not in used_objects and object.name.split("_")[0] != self.object.name.split("_")[0]:
                self.secondary_objects.append(object)
                used_objects.add(object.name)
                scene_xml = object.add_to_scene_xml(scene_xml)
        self.texture = self.textures_manager.sample(random=True)
        scene_xml = include_in_scene(scene_xml, os.path.join(assets_dir, "embodiments", self.robot.name, "model_pick.xml"))
        scene_xml = scene_xml.replace("{ASSETS_DIR}", assets_dir)
        scene_xml = scene_xml.replace("{TEXTURE_PATH}", self.texture)
        scene_xml = scene_xml.replace("{TABLE_WIDTH}", str(self.spread+0.24))
        scene_xml = scene_xml.replace("{TABLE_HEIGHT}", str(self.spread+0.24))
        return scene_xml

    def get_obs(self):
        obs = self.get_base_obs()
        obs["object_pose"] = self.object.get_pose(self.data).reshape(16).astype(np.float32)
        obs["object_name"] = self.object.name.split("_")[0]
        return obs

    def compute_reward(self) -> float:
        object_pose = self.object.get_pose(self.data)
        lift_distance = object_pose[2][3] - self.object.last_set_pose[2][3]
        return lift_distance

    def reset(self, seed=None, options=None):
        super().reset(seed=seed, options=options)
        scene_xml_string = self.make_task_scene()
        self.grasped_bodies = set()
        self.grasping_object = False
        
        gripper_init_pose = np.eye(4)
        gripper_init_pose[:3, 3] = np.array([0.00, -0.68, 1.1])
        gripper_init_pose[:3, :3] = R.from_euler("xyz", np.array([1.3, 0.000, 0.000])).as_matrix()
        self.setup_mujoco(scene_xml_string)
        initial_positions = position_sampler(self.np_random, len(self.secondary_objects) + 1, [self.spread,self.spread], self.spread/2, 100, start_position=gripper_init_pose[:2, 3], thickness=self.spread/2.5)
        pose = np.eye(4)
        self.initial_robot_pose = self.robot.get_camera_pose()
        self.env_step(10)

        for i, obj in enumerate([self.object] + self.secondary_objects):
            z = obj.get_center_to_bottom_z_distance(self.data) + 0.78
            pose[:2, 3] = initial_positions[i][:2]
            pose[2, 3] = z
            pose[:3, :3] = R.from_euler(
                "x", self.np_random.uniform(0, 2 * np.pi)
            ).as_matrix()
            obj.set_pose(self.model, self.data, pose)
        
        self.env_step(100)
        self.settle_env()
        self.env_step(100)
        for obj in [self.object] + self.secondary_objects:
            obj.last_set_pose = obj.get_pose(self.data).copy()
        
        self.robot.prepare(gripper_init_pose, self.env_step)
        self.env_step(100)        
        self.enable_sleeping_islands()

        observation = self.get_obs()
        info = {"object_name": self.object.name}

        return observation, info

    def step(self, action):
        obs, reward, terminated, truncated, info = super().step(action)
        self.grasped_bodies.update(self.robot.get_grasped_bodies())
        grasped_bodies_list = list(self.grasped_bodies)
        if f"{self.object.name}_object" in grasped_bodies_list:
            self.grasping_object = True
        info = {
            "grasped_bodies": grasped_bodies_list,
            "object_name": self.object.name,
            "texture_name": self.texture,
            "initial_robot_pose": self.initial_robot_pose,
            "initial_object_pose": self.object.last_set_pose,
            "is_grasping": self.robot.get_grasp(),
            "gripper_current_position": self.robot.get_tcp_pose(),
            "grasping_object": self.grasping_object,
            **info
        }

        if np.linalg.norm(self.robot.get_camera_pose()[:3,3] - self.object.get_pose(self.data)[:3,3]) > np.linalg.norm(self.initial_robot_pose[:3,3] - self.object.last_set_pose[:3,3])*1.5:
            truncated = True
            terminated = True

        return obs, reward, terminated, truncated, info



import torch
import torch.nn as nn

class CustomModel(nn.Module):
    def __init__(self, *args, **kwargs):
        super().__init__()
        # TODO: set model
        self.model = None

    def to(self, device):
        # TODO: move model to device
        if self.model is not None:
            self.model.to(device)
        return self

    def forward(self, x):
        # TODO: forward pass of model
        pass
    
    def step(self, data, *args, **kwargs):
        images, actions = data
        # images is a tensor of shape (1, image_buffer_size, 3, 256, 256), where the images are in chronological order
        # actions is a tensor of shape (1, image_buffer_size, 7), where the actions are in chronological order, and the final action is padding
        
        # TODO: implement pass that takes in observations as described above out outputs a 7-dimensional action
        action = torch.zeros(7)
        logs = {}
        
        return action, logs

    def reset(self):
        # TODO: optional; this method is called once the robot is homed
        pass


import zmq
import time
import pickle
import threading
import traceback


class RPCServer:
    def __init__(self, obj, port: int = 5000, threaded=False):
        """
        obj: object with methods to expose
        port: port to listen on
        """
        self.obj = obj
        self.context = zmq.Context()
        self.socket: zmq.socket.Socket = self.context.socket(zmq.REP)
        self.socket.bind(f"tcp://*:{port}")
        self.threaded = threaded
        if threaded:
            self.thread = threading.Thread(target=self.run)
            self.stop_event = threading.Event()
        else:
            self.stop_event = False

    def _send_exception(self, e):
        """
        Serialize an exception and send it over the socket.
        Only the exception type, message, and traceback are sent.
        """
        exception = {
            "type": "exception",
            "content": {
                "exception": str(type(e)),
                "message": str(e),
                "traceback": traceback.format_exc(),
            },
        }
        self.socket.send(pickle.dumps(exception))

    def _send_result(self, result):
        """
        Serialize a result and send it over the socket.
        """
        result = {"type": "result", "content": result}
        self.socket.send(pickle.dumps(result))

    def run(self):
        """
        Run the server.
        """
        if self.threaded:
            while not self.stop_event.is_set():
                message = self.socket.recv()
                message = pickle.loads(message)
                self._handle_message(message)
        else:
            while not self.stop_event:
                try:
                    message = self.socket.recv(flags=zmq.NOBLOCK)
                    message = pickle.loads(message)
                except zmq.Again:
                    time.sleep(0.001)
                    continue
                self._handle_message(message)

    def _is_callable(self, attr):
        return hasattr(self.obj, attr) and callable(getattr(self.obj, attr))

    def _handle_message(self, message):
        """
        Handles a dictionary of {
            "req": str,  # request type
            "attr": str,
            "args": list,
            "kwargs": dict,
        }
        from the socket.
        If req == "is_callable", return whether the attribute is callable.
        If req == "get", return the attribute.
            If the attribute is not found, return an error message.
            If the attribute is callable, call with args and kwargs.
                If there are any errors in the callable, return the pickled error
                If the callable is found and there are no errors, return the pickled result.
            If the attribute is not callable, return the attribute.
        If req == "set", set the attribute to the value.
        If req == "dir", return a list of attributes.
        If req == "stop", stop the server.
        """
        if message["req"] == "is_callable":
            result = self._is_callable(message["attr"])
            self._send_result(result)
        elif message["req"] == "get":
            try:
                attribute = getattr(self.obj, message["attr"])
                args = message["args"]
                kwargs = message["kwargs"]
                if not callable(attribute):
                    self._send_result(attribute)
                else:
                    result = attribute(*args, **kwargs)
                    self._send_result(result)
            except Exception as e:
                self._send_exception(e)
        elif message["req"] == "set":
            try:
                setattr(self.obj, message["attr"], message["value"])
                self._send_result(None)
            except Exception as e:
                self._send_exception(e)
        elif message["req"] == "dir":
            result = dir(self.obj)
            self._send_result(result)
        elif message["req"] == "stop":
            self.stop()

    def close(self):
        self.socket.close()
        self.context.term()

    def start(self):
        if self.threaded:
            self.stop_event.clear()
            self.thread.start()
        else:
            self.run()

    def stop(self):
        if self.threaded:
            self.stop_event.set()
            self.thread.join()
        else:
            self.stop_event = True
        self.close()


class RPCException(Exception):
    def __init__(self, exception_type: str, message: str, traceback: str):
        self.exception_type = exception_type
        self.message = message
        self.traceback = traceback

    def __str__(self):
        return f"{self.exception_type}: {self.message}\n{self.traceback}"


class RPCClient:
    def __init__(self, host: str, port: int = 5000):
        """
        host: host to connect to
        port: port to connect to
        """
        self.__dict__["context"] = zmq.Context()
        self.__dict__["socket"] = self.context.socket(zmq.REQ)
        self.socket.connect(f"tcp://{host}:{port}")
        self.__dict__["_is_callable_cache"] = {}

    def __setattr__(self, attr: str, value):
        """
        Set the attribute of the same name.
        Attribute must not be callable on remote.
        """
        if self._is_callable(attr):
            raise AttributeError(f"Overwriting a callable attribute: {attr}")
        self._send_set(attr, value)

    def _send_get(self, attr: str, args: list, kwargs: dict):
        """
        Send a get request over the socket.
        """
        req = {"req": "get", "attr": attr, "args": args, "kwargs": kwargs}
        self.socket.send(pickle.dumps(req))
        return self._recv_result()

    def _send_set(self, attr: str, value):
        """
        Send a set request over the socket.
        """
        req = {"req": "set", "attr": attr, "value": value}
        self.socket.send(pickle.dumps(req))
        return self._recv_result()

    def _recv_result(self):
        """
        Receive a dictionary of {
            "type": str,
            "content": object,
        }
        if type == "exception", content is a dictionary of {
            "exception": str,
            "message": str,
            "traceback": str,
        }; re-raise the exception on the client side
        if type == "result", content is the result
        """
        result = self.socket.recv()
        result = pickle.loads(result)
        if result["type"] == "exception":
            raise RPCException(
                result["content"]["exception"],
                result["content"]["message"],
                result["content"]["traceback"],
            )
        return result["content"]

    def _is_callable(self, attr: str) -> bool:
        """
        Send a request to check if the attribute is callable.
        Returns False if the attribute is not found.
        """
        if attr not in self._is_callable_cache:
            req = {"req": "is_callable", "attr": attr}
            self.socket.send(pickle.dumps(req))
            result = self._recv_result()
            self._is_callable_cache[attr] = result
        return self._is_callable_cache[attr]

    def __getattr__(self, attr: str):
        """
        Return the attribute of the same name.
        If the attribute is a callable, return a function that sends the call over the socket.
        Else, return the attribute value.
        """
        if self._is_callable(attr):
            return lambda *args, **kwargs: self._send_get(attr, args, kwargs)
        else:
            return self._send_get(attr, [], {})

    def __dir__(self):
        """
        Return a list of attributes.
        """
        req = {"req": "dir"}
        self.socket.send(pickle.dumps(req))
        result = self._recv_result()
        return result + ["stop_server"]

    def stop_server(self) -> bool:
        """
        Send a stop request to the server.
        If the server is stopped, close the socket and terminate the context.
        Returns a bool for success.
        """
        req = {"req": "stop"}
        self.socket.send(pickle.dumps(req))
        stopped = self._recv_result()
        if stopped:
            self.socket.close()
            self.context.term()
        return stopped



from baselines.rum.models.bet.vqvae.vector_quantize_pytorch import VectorQuantize
from baselines.rum.models.bet.vqvae.residual_vq import ResidualVQ
from baselines.rum.models.bet.vqvae.vqvae import VqVae



import logging
from enum import Enum
from itertools import chain
from typing import Dict, Optional, Sequence, Tuple, Union

import accelerate
import einops
import torch
import torch.nn as nn
import torch.nn.functional as F

from baselines.rum.models.bet.bet import FocalLoss, KMeansDiscretizer
from baselines.rum.models.bet.gpt import GPT
from baselines.rum.models.bet.utils import MLP

GENERATOR_SEED_FIXED = 123456789


class TokenizedBehaviorTransformer(nn.Module):
    GOAL_SPEC = Enum("GOAL_SPEC", "concat stack unconditional")

    def __init__(
        self,
        obs_dim: int,
        act_dim: int,
        goal_dim: int,
        gpt_model: GPT,
        action_spec: Optional[Sequence[int]] = None,
        start_and_ends: Optional[Sequence[Tuple[int, int]]] = None,
        num_extra_predicted_actions: Optional[int] = None,
        n_clusters: Union[int, Sequence[int]] = 32,
        kmeans_fit_steps: int = 500,
        kmeans_iters: int = 50,
        offset_loss_multiplier: float = 1.0e3,
        offset_distance_metric: str = "L2",
        representation_height: int = 7,
        representation_width: int = 7,
        gamma: float = 2.0,
        sampling_temperature: float = 1.0,
        **kwargs,
    ):
        super().__init__()
        self._obs_dim = obs_dim
        self._act_dim = act_dim
        self.sampling_temperature = sampling_temperature
        # First ensure either action spec is given or start and ends are given.
        assert (action_spec is not None) != (start_and_ends is not None), (
            "EITHER action_spec OR start_and_ends must be given."
        )
        if action_spec is not None:
            self._action_spec = action_spec
            assert act_dim == sum(action_spec)
            cumsum = [sum(action_spec[:i]) for i in range(len(action_spec) + 1)]
            self._starts = cumsum[:-1]
            self._ends = cumsum[1:]
            self._start_ends = zip(self._starts, self._ends)
        else:
            self._start_ends = start_and_ends
            self._action_spec = [end - start for (start, end) in start_and_ends]
        self._start_and_ends = list(enumerate(self._start_ends))
        self._n_subactions = len(self._start_and_ends)  # Number of sub-actions.

        self._goal_dim = goal_dim
        self._num_extra_predicted_actions = num_extra_predicted_actions

        # Decide goal conditioning style.
        if goal_dim <= 0:
            self._cbet_method = self.GOAL_SPEC.unconditional
        elif obs_dim == goal_dim:
            self._cbet_method = self.GOAL_SPEC.concat
        else:
            self._goal_encoder = nn.Linear(goal_dim, obs_dim, bias=False)
            self._cbet_method = self.GOAL_SPEC.stack

        self._gpt_model = gpt_model
        if isinstance(n_clusters, int):
            n_clusters = [n_clusters] * len(self._start_and_ends)
        assert len(n_clusters) == len(self._start_and_ends)
        # For now, we assume the number of clusters is given.
        assert all(k > 0 for k in n_clusters)
        self._K = n_clusters
        self._kmeans_fit_steps = kmeans_fit_steps
        self._clustering_algos = [
            KMeansDiscretizer(num_bins=k, kmeans_iters=kmeans_iters) for k in n_clusters
        ]
        self._current_steps = 0
        self._map_to_cbet_preds = nn.ModuleList(
            [
                MLP(
                    in_channels=gpt_model.config.output_dim,
                    hidden_channels=[(a + 1) * k],
                )
                for (a, k) in zip(self._action_spec, n_clusters)
            ]
        )
        self._collected_actions = []
        self._have_fit_kmeans = False
        # Placeholder for the cluster centers.
        generator = torch.Generator()
        generator.manual_seed(GENERATOR_SEED_FIXED)
        self._cluster_centers = nn.ParameterList(
            [
                nn.Parameter(
                    torch.randn((k, a), generator=generator, dtype=torch.float32),
                    requires_grad=False,
                )
                for (k, a) in zip(n_clusters, self._action_spec)
            ]
        )
        self._criterion = FocalLoss(gamma=gamma, reduction="none")
        self._offset_criterion = (
            nn.MSELoss(reduction="none")
            if offset_distance_metric == "L2"
            else nn.L1Loss(reduction="none")
        )
        self._offset_loss_multiplier = offset_loss_multiplier

        self._action_tokenizer = nn.ModuleList(
            [nn.Linear(a, obs_dim) for a in self._action_spec]
        )
        # Figure out the embedding tokens.
        if self._cbet_method == self.GOAL_SPEC.unconditional:
            self._goal_embedding_token = None
        else:
            self._goal_embedding_token = nn.Parameter(torch.randn(goal_dim))
        self._h, self._w = representation_height, representation_width
        self._obs_embedding_token = nn.Parameter(
            torch.randn([obs_dim, self._h, self._w])
        )
        self._action_embedding_token = nn.Parameter(
            torch.randn([self._n_subactions, obs_dim])
        )
        self._extra_action_token = nn.Parameter(
            torch.randn([self._n_subactions, obs_dim])
        )
        self._end_of_obs_token = nn.Parameter(torch.randn(obs_dim))
        self._accelerator = accelerate.Accelerator()

    def _load_from_state_dict(self, *args, **kwargs):
        # Don't fit kmeans if we are loading from a state dict.
        self._current_steps = self._kmeans_fit_steps
        self._have_fit_kmeans = True
        return super()._load_from_state_dict(*args, **kwargs)

    def get_start_and_ends(self) -> Sequence[Tuple[int, int]]:
        return self._start_and_ends

    def _tokenize_actions(self, actions: torch.Tensor) -> torch.Tensor:
        action_stack = [
            self._action_tokenizer[i](actions[..., start:end])
            for i, (start, end) in self._start_and_ends
        ]
        return torch.stack(action_stack, dim=2)

    def _detokenize_actions(
        self, tokenized_action_output: torch.Tensor
    ) -> torch.Tensor:
        action_bin_logits = []
        all_action_offsets = []
        for i, _ in self._start_and_ends:
            tokenized_action_i = tokenized_action_output[..., i, :]
            action_cbet_preds = self._map_to_cbet_preds[i](tokenized_action_i)
            action_center_logits, action_offsets = torch.split(
                action_cbet_preds,
                [self._K[i], self._K[i] * self._action_spec[i]],
                dim=-1,
            )
            action_bin_logits.append(action_center_logits)
            all_action_offsets.append(action_offsets)
        return action_bin_logits, all_action_offsets

    def _begin_epoch(self, optimizer, **kwargs):
        # log learning rate for debugging
        lr_0 = optimizer.param_groups[0]["lr"]
        lr_neg1 = optimizer.param_groups[-1]["lr"]
        return {"lr_0": lr_0, "lr_neg1": lr_neg1}

    def _calculate_cbet_preds_and_loss(
        self,
        cluster_centers: torch.Tensor,
        bin_logits: torch.Tensor,
        action_offsets: torch.Tensor,
        true_actions: torch.Tensor,
        is_padded_action_seq: Optional[torch.Tensor],
        predict_with_offset: bool = True,
        return_loss: bool = True,
        sampling_temperature: float = 1.0,
    ) -> torch.Tensor:
        bin_probs = torch.softmax(bin_logits / sampling_temperature, dim=-1)
        N, T, choices = bin_probs.shape
        # Sample from the multinomial distribution, one per row.
        sampled_centers = einops.rearrange(
            torch.multinomial(bin_probs.view(-1, choices), num_samples=1),
            "(N T) 1 -> N T 1",
            N=N,
        )
        flattened_action_offsets = einops.rearrange(
            action_offsets, "N T (K A) -> (N T) K A", K=choices
        )
        sampled_offsets = flattened_action_offsets[
            torch.arange(flattened_action_offsets.shape[0]), sampled_centers.flatten()
        ].view(N, T, -1)
        centers = cluster_centers[sampled_centers.flatten()].view(N, T, -1)
        a_hat = centers + (sampled_offsets if predict_with_offset else 0.0)
        if not return_loss:
            return a_hat, None, {}
        # We are in training, so figure out the loss for the actions.
        # First, we need to find the closest cluster center for each action.
        true_action_bins = self._find_closest_cluster(true_actions, cluster_centers)
        true_offsets = true_actions - cluster_centers[true_action_bins]
        predicted_offsets = flattened_action_offsets[
            torch.arange(flattened_action_offsets.shape[0]), true_action_bins.flatten()
        ].view(N, T, -1)
        offset_loss = self._offset_criterion(predicted_offsets, true_offsets)
        cbet_loss = self._criterion(
            einops.rearrange(bin_logits, "N T D -> (N T) D"),
            einops.rearrange(true_action_bins, "N T -> (N T)"),
        )
        if is_padded_action_seq is not None:
            cbet_loss *= ~is_padded_action_seq.view(-1)
            offset_loss *= ~is_padded_action_seq.unsqueeze(-1)
        cbet_loss, offset_loss = cbet_loss.mean(), offset_loss.mean()
        action_mse = F.mse_loss(a_hat, true_actions, reduction="none")
        action_l1 = F.l1_loss(a_hat, true_actions, reduction="none")
        norm = torch.norm(true_actions, p=2, dim=-1, keepdim=True) + 1e-9
        normalized_mse = (action_mse / norm).mean()
        loss = cbet_loss + self._offset_loss_multiplier * offset_loss
        if self._current_steps < self._kmeans_fit_steps:
            loss = loss.detach() + (loss * 0.0)
        loss_dict = {
            "classification_loss": cbet_loss.detach().cpu().item(),
            "offset_loss": offset_loss.detach().cpu().item(),
            "loss": loss.detach().cpu().item(),
            "L2_loss": action_mse.mean().detach().cpu().item(),
            "L2_loss_normalized": normalized_mse.mean().detach().cpu().item(),
            "L1_loss": action_l1.mean().detach().cpu().item(),
        }
        return a_hat, loss, loss_dict

    def forward(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: Optional[torch.Tensor],
        padding_seq: Optional[torch.Tensor],
        predict_with_offset: bool = True,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
        if self._current_steps == 0:
            self._cluster_centers = self._cluster_centers.to(obs_seq.device)
        if self._current_steps < self._kmeans_fit_steps and (
            action_seq is not None and padding_seq is not None
        ):
            self._current_steps += 1
            self._fit_kmeans(action_seq, padding_seq)
        return self._predict(
            obs_seq,
            goal_seq,
            action_seq,
            padding_seq,
            predict_with_offset=predict_with_offset,
        )

    def _fit_kmeans(
        self,
        action_seq: Optional[torch.Tensor],
        padding_seq: Optional[torch.Tensor],
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
        assert self._current_steps <= self._kmeans_fit_steps
        if self._current_steps == 1:
            self._cluster_centers = self._cluster_centers.to(action_seq.device)

        all_action_seq = self._accelerator.gather(action_seq)
        all_padding_seq = self._accelerator.gather(padding_seq)
        self._collected_actions.append(
            all_action_seq[torch.logical_not(all_padding_seq)]
        )
        if self._current_steps == self._kmeans_fit_steps:
            logging.info("Fitting KMeans")
            self._collected_actions = torch.cat(self._collected_actions, dim=0)
            for i, (start, end) in self._start_and_ends:
                clustering_algo = self._clustering_algos[i]
                logging.info(f"Fitting KMeans for action {i}")
                clustering_algo.fit(
                    self._collected_actions[:, start:end].view(-1, end - start)
                )
                self._cluster_centers[i] = clustering_algo.bin_centers.float().to(
                    action_seq.device
                )
            self._have_fit_kmeans = True

    def _predict(
        self,
        obs_seq: torch.Tensor,
        goal_seq: Optional[torch.Tensor],
        action_seq: torch.Tensor,
        is_padded_action_seq: Optional[torch.Tensor],
        predict_with_offset: bool = True,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Dict[str, float]]:
        batch_size, obs_T, c, h, w = obs_seq.shape
        obs_seq_and_encoding = obs_seq + self._obs_embedding_token[None, None, ...]
        obs_seq_and_encoding = einops.rearrange(
            obs_seq_and_encoding, "N T C H W -> N T (H W) C"
        )
        assert action_seq is not None
        tokenized_action_seq = self._tokenize_actions(action_seq)
        tokenized_action_seq_and_encoding = (
            tokenized_action_seq + self._action_embedding_token[None, None, ...]
        )
        interspersed_seq = torch.cat(
            [obs_seq_and_encoding, tokenized_action_seq_and_encoding[:, :obs_T, ...]],
            dim=2,
        )  # N T (H W + A) C
        interspersed_seq_flattened = einops.rearrange(
            interspersed_seq, "N T (HWA) C -> N (T HWA) C"
        )
        extra_actions_tokenized_plus_embedding = (
            tokenized_action_seq_and_encoding[:, obs_T:, ...]
            + self._extra_action_token[None, None, ...]
        )
        extra_actions_flattened = einops.rearrange(
            extra_actions_tokenized_plus_embedding, "N T A D -> N (T A) D"
        )
        interspersed_seq_and_extra_actions = torch.cat(
            [
                interspersed_seq_flattened,
                einops.repeat(self._end_of_obs_token, "D -> N 1 D", N=batch_size),
                extra_actions_flattened,
            ],
            dim=1,
        )

        # Assume dimensions are N T D for N sequences of T timesteps with dimension D.
        if self._cbet_method == self.GOAL_SPEC.unconditional:
            gpt_input = interspersed_seq_and_extra_actions
        elif self._cbet_method == self.GOAL_SPEC.concat:
            goal_seq_encoded = goal_seq + self._goal_embedding_token[None, None, ...]
            gpt_input = torch.cat(
                [goal_seq_encoded, interspersed_seq_and_extra_actions], dim=1
            )
        elif self._cbet_method == self.GOAL_SPEC.stack:
            goal_seq_embedded = (
                goal_seq[:, None, :] + self._goal_embedding_token[None, None, ...]
            )
            goal_seq_encoded = self._goal_encoder(goal_seq_embedded)
            gpt_input = torch.cat(
                [goal_seq_encoded, interspersed_seq_and_extra_actions], dim=1
            )
        else:
            raise NotImplementedError

        gpt_output = self._gpt_model(gpt_input)
        if self._cbet_method == self.GOAL_SPEC.concat:
            # Chop off the goal encodings.
            gpt_output = gpt_output[:, goal_seq.size(1) :, :]
        elif self._cbet_method == self.GOAL_SPEC.stack:
            gpt_output = gpt_output[:, 1::]

        # Here we have a sequence of shape (N, (T*(H*W + A) + 1 + T'*A), D)
        # where T' is the number of extra actions we want to predict.
        # Separate out the original and the extra actions.
        # We have H*W + A tokens per obs, + 1 for the end of obs token.
        extra_action_output_flat = gpt_output[
            :, obs_T * (self._h * self._w + self._n_subactions) : -1, :
        ]  # N (T' A) D
        extra_action_output = einops.rearrange(
            extra_action_output_flat, "N (T A) D -> N T A D", A=self._n_subactions
        )  # N T' A D
        original_output_tokens = gpt_output[
            :, : obs_T * (self._h * self._w + self._n_subactions), :
        ]
        original_output_tokens_reshaped = einops.rearrange(
            original_output_tokens, "N (T HWA) D -> N T HWA D", T=obs_T
        )
        original_action_tokens = original_output_tokens_reshaped[
            :, :, -self._n_subactions - 1 : -1, :
        ]  # N T A D
        output_action_tokens = torch.cat(
            [original_action_tokens, extra_action_output], dim=1
        )  # N (T + T') A D
        action_bin_logits, action_offsets = self._detokenize_actions(
            output_action_tokens
        )
        # Now calculate the predicted actions and the losses.
        a_hat, loss, loss_dict = torch.zeros_like(action_seq), None, {}
        for i, (start, end) in self._start_and_ends:
            a_hat_i, loss_i, loss_dict_i = self._calculate_cbet_preds_and_loss(
                self._cluster_centers[i],
                action_bin_logits[i],
                action_offsets[i],
                action_seq[..., start:end],
                # TODO make is_padded_action_seq more flexible,
                # For example, in reality part of the action could be "padded"
                is_padded_action_seq,
                predict_with_offset=predict_with_offset,
                return_loss=True,
                sampling_temperature=self.sampling_temperature,
            )
            a_hat[..., start:end] = a_hat_i
            if i == 0:
                # a_hat = a_hat_i
                loss = loss_i
                for k, v in loss_dict_i.items():
                    loss_dict[k] = v
            else:
                # a_hat = torch.cat([a_hat, a_hat_i], dim=-1)
                loss += loss_i
                for k, v in loss_dict_i.items():
                    loss_dict[k] += v
            for k, v in loss_dict_i.items():
                loss_dict[f"{k}_{i}"] = v
        return a_hat, loss, loss_dict

    def _find_closest_cluster(
        self, action_seq: torch.Tensor, cluster_centers: torch.Tensor
    ) -> torch.Tensor:
        N, T, _ = action_seq.shape
        flattened_actions = einops.rearrange(action_seq, "N T A -> (N T) A")
        cluster_center_distance = torch.sum(
            (flattened_actions[:, None, :] - cluster_centers[None, :, :]) ** 2,
            dim=2,
        )  # (N T) K A -> (N T) K
        closest_cluster_center = torch.argmin(cluster_center_distance, dim=1)  # (N T)
        discretized_action = einops.rearrange(
            closest_cluster_center, "(N T) -> N T", N=N, T=T
        )
        return discretized_action

    def configure_optimizers(self, weight_decay, learning_rate, betas):
        optimizer = self._gpt_model.configure_optimizers(
            weight_decay=weight_decay,
            learning_rate=learning_rate,
            betas=betas,
        )
        optimizer.add_param_group(
            {
                "params": chain(
                    self._map_to_cbet_preds.parameters(),
                    self._action_tokenizer.parameters(),
                )
            }
        )
        optimizer.add_param_group(
            {
                "params": [
                    self._obs_embedding_token,
                    self._goal_embedding_token,
                    self._action_embedding_token,
                    self._extra_action_token,
                    self._end_of_obs_token,
                ],
                "weight_decay": 0.0,
            }
        )
        return optimizer
